# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import re
import torch
import numpy as np
import pandas as pd
import hashlib
from PIL import Image, ImageDraw
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.utils import pop_value, nested_dict_value
from unitorch.models.dpt import (
    DPTForDepthEstimation as _DPTForDepthEstimation,
    DPTProcessor,
)
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli import CoreConfigureParser, GenericScript
from unitorch.cli import register_script
from unitorch.cli.models.dpt import pretrained_dpt_infos


class DPTForDepthEstimationPipeline(_DPTForDepthEstimation):
    def __init__(
        self,
        config_path: str,
        vision_config_path: str,
        weight_path: Optional[Union[str, List[str]]] = None,
        state_dict: Optional[Dict[str, Any]] = None,
        enable_cpu_offload: Optional[bool] = True,
        device: Optional[Union[str, int]] = "cpu",
    ):
        super().__init__(
            config_path=config_path,
        )
        self.processor = DPTProcessor(
            vision_config_path=vision_config_path,
        )
        self._device = "cpu" if device == "cpu" else int(device)

        self.from_pretrained(weight_path, state_dict=state_dict)
        self._enable_cpu_offload = enable_cpu_offload
        if not self._enable_cpu_offload and self._device != "cpu":
            self.to(device=self._device)
        self.eval()

    @classmethod
    @add_default_section_for_init("core/pipeline/dpt")
    def from_core_configure(
        cls,
        config,
        pretrained_name: Optional[str] = None,
        config_path: Optional[str] = None,
        vision_config_path: Optional[str] = None,
        pretrained_weight_path: Optional[str] = None,
        device: Optional[str] = None,
        **kwargs,
    ):
        config.set_default_section("core/pipeline/dpt")

        pretrained_name = pretrained_name or config.getoption(
            "pretrained_name", "dpt-large"
        )
        config_path = config_path or config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_dpt_infos, pretrained_name, "config"),
        )
        config_path = cached_path(config_path)

        vision_config_path = vision_config_path or config.getoption(
            "vision_config_path", None
        )
        vision_config_path = pop_value(
            vision_config_path,
            nested_dict_value(pretrained_dpt_infos, pretrained_name, "vision_config"),
        )
        vision_config_path = cached_path(vision_config_path)

        enable_cpu_offload = config.getoption("enable_cpu_offload", True)
        device = config.getoption("device", "cpu") if device is None else device
        pretrained_weight_path = pretrained_weight_path or config.getoption(
            "pretrained_weight_path", None
        )
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_dpt_infos, pretrained_name, "weight"),
            check_none=False,
        )

        inst = cls(
            config_path=config_path,
            vision_config_path=vision_config_path,
            weight_path=weight_path,
            enable_cpu_offload=enable_cpu_offload,
            device=device,
        )

        return inst

    @torch.no_grad()
    @add_default_section_for_function("core/pipeline/dpt")
    def __call__(
        self,
        image: Union[Image.Image, str],
    ):
        if self._enable_cpu_offload:
            self.to(self._device)
        inputs = self.processor.classification(image)
        pixel_values = inputs.pixel_values.unsqueeze(0).to(self._device)
        outputs = self.forward(
            pixel_values,
        )
        masks = outputs[0].cpu().numpy().squeeze(0)
        result_image = Image.fromarray((masks * 255) / np.max(masks))
        result_image = result_image.resize(image.size, resample=Image.LANCZOS)
        if self._enable_cpu_offload:
            self.to("cpu")
            torch.cuda.empty_cache()

        return result_image



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli import hf_endpoint_url

# pretrained infos
pretrained_swin_infos = {
    "swin-tiny-patch4-window7-224": {
        "config": hf_endpoint_url(
            "/microsoft/swin-tiny-patch4-window7-224/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/microsoft/swin-tiny-patch4-window7-224/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/microsoft/swin-tiny-patch4-window7-224/resolve/main/pytorch_model.bin"
        ),
    },
    "swin-base-patch4-window7-224": {
        "config": hf_endpoint_url(
            "/microsoft/swin-base-patch4-window7-224/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/microsoft/swin-base-patch4-window7-224/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/microsoft/swin-base-patch4-window7-224/resolve/main/pytorch_model.bin"
        ),
    },
    "swin-large-patch4-window7-224": {
        "config": hf_endpoint_url(
            "/microsoft/swin-large-patch4-window7-224/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/microsoft/swin-large-patch4-window7-224/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/microsoft/swin-large-patch4-window7-224/resolve/main/pytorch_model.bin"
        ),
    },
    "swin-large-patch4-window12-384": {
        "config": hf_endpoint_url(
            "/microsoft/swin-large-patch4-window12-384/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/microsoft/swin-large-patch4-window12-384/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/microsoft/swin-large-patch4-window12-384/resolve/main/pytorch_model.bin"
        ),
    },
    "swin-base-patch4-window7-224-in22k": {
        "config": hf_endpoint_url(
            "/microsoft/swin-base-patch4-window7-224-in22k/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/microsoft/swin-base-patch4-window7-224-in22k/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/microsoft/swin-base-patch4-window7-224-in22k/resolve/main/pytorch_model.bin"
        ),
    },
    "swin-large-patch4-window7-224-in22k": {
        "config": hf_endpoint_url(
            "/microsoft/swin-large-patch4-window7-224-in22k/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/microsoft/swin-large-patch4-window7-224-in22k/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/microsoft/swin-large-patch4-window7-224-in22k/resolve/main/pytorch_model.bin"
        ),
    },
}

import unitorch.cli.models.swin.modeling
import unitorch.cli.models.swin.processing
from unitorch.cli.models.swin.modeling import SwinForImageClassification
from unitorch.cli.models.swin.processing import SwinProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import torch
import time
import json
import logging
import shutil
import numpy as np
import pandas as pd
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
from copy import deepcopy
from itertools import chain
from collections.abc import Iterable
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union, Iterator
from torch.utils.data import DataLoader, Dataset, SequentialSampler
from torch.utils.data.distributed import DistributedSampler
from torch.multiprocessing import Process, Queue
from megatron.core import mpu, dist_checkpointing, tensor_parallel
from megatron.core.pipeline_parallel.schedules import (
    get_forward_backward_func,
    forward_backward_no_pipelining,
    forward_backward_pipelining_with_interleaving,
    forward_backward_pipelining_without_interleaving,
)
from megatron.core.optimizer import OptimizerConfig, get_megatron_optimizer
from megatron.core.distributed import (
    DistributedDataParallelConfig,
    DistributedDataParallel,
)
from megatron.core.optimizer import DistributedOptimizer
from unitorch import set_seed
from unitorch.models import ExponentialMovingAverage
from unitorch.utils import (
    get_local_rank,
    nested_dict_value,
    update_nested_dict,
)
from unitorch.utils import (
    DistributedSkipSampler,
    RandomSkipSampler,
    SequentialSkipSampler,
    PostProcess,
    IOProcess,
    GENERATE_FINISHED,
)
from unitorch.models import GenericOutputs
from unitorch.cli import (
    cached_path,
    register_task,
    registered_model,
    registered_optim,
    registered_dataset,
    registered_loss,
    registered_score,
    registered_scheduler,
    registered_writer,
    init_registered_module,
    init_registered_process,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli.models import (
    ModelInputs,
    ModelOutputs,
    ModelTargets,
    LossOutputs,
    CombineTensorsInputs,
    CombineTensorsTargets,
)
from unitorch.cli.tasks.supervised import collate_fn
import unitorch.cli.wandb as wandb


def get_batch_on_this_cp_rank(batch: Dict[str, Any]):
    cp_size = mpu.get_context_parallel_world_size()
    if cp_size > 1:
        cp_rank = mpu.get_context_parallel_rank()
        for key, val in batch.items():
            if val is not None:
                seq_dim = 1  # if key != 'attention_mask' else 2
                val = val.view(
                    *val.shape[0:seq_dim],
                    2 * cp_size,
                    val.shape[seq_dim] // (2 * cp_size),
                    *val.shape[(seq_dim + 1) :],
                )
                index = torch.tensor(
                    [cp_rank, (2 * cp_size - cp_rank - 1)],
                    device="cpu",
                    pin_memory=True,
                ).cuda(non_blocking=True)
                val = val.index_select(seq_dim, index)
                val = val.view(*val.shape[0:seq_dim], -1, *val.shape[(seq_dim + 2) :])
                batch[key] = val

    return batch


def process_batch_data(inputs, targets):
    data = [inputs, targets]
    dist.broadcast_object_list(
        data,
        src=mpu.get_tensor_model_parallel_src_rank(),
        group=mpu.get_tensor_model_parallel_group(),
    )
    inputs, targets = data[0], data[1]

    batch = {**inputs, **targets}
    batch = {k: v.cuda() if v is not None else v for k, v in batch.items()}
    batch = get_batch_on_this_cp_rank(batch)
    return batch


@register_task("core/task/megatron/supervised")
class MegatronTask:
    """Task class for deepspeed supervised learning."""

    def __init__(
        self,
        configure,
        model,
        datasets,
        local_rank: Optional[int] = -1,
        seed: Optional[int] = 1123,
    ):
        """
        Initialize the DeepspeedTask.

        Args:
            configure: The configuration object.
            model: The model for supervised learning.
            datasets: The datasets for training and evaluation.
            local_rank (optional): The local rank for distributed training. Defaults to -1.
            seed (optional): The random seed. Defaults to 1123.
        """
        set_seed(seed)
        tensor_parallel.model_parallel_cuda_manual_seed(seed)
        self.config = configure
        self.model = model
        self.datasets = datasets
        self.local_rank = local_rank

        if self.local_rank != -1:
            torch.cuda.set_device(self.local_rank)

        if torch.cuda.is_available():
            self.model = self.model.cuda()

        self.best_score = -np.inf
        self.dp_rank = mpu.get_data_parallel_rank()
        self.dp_size = mpu.get_data_parallel_world_size()
        self.pp_rank = mpu.get_pipeline_model_parallel_rank()
        self.pp_size = mpu.get_pipeline_model_parallel_world_size()
        self.vp_rank = mpu.get_virtual_pipeline_model_parallel_rank()
        self.vp_size = mpu.get_virtual_pipeline_model_parallel_world_size()
        self.cp_rank = mpu.get_context_parallel_rank()
        self.cp_group = mpu.get_context_parallel_group()
        self.cp_size = mpu.get_context_parallel_world_size()
        self.tp_rank = mpu.get_tensor_model_parallel_rank()
        self.tp_size = mpu.get_tensor_model_parallel_world_size()
        self.is_pp_last_rank = mpu.is_pipeline_last_stage(ignore_virtual=True)

    @classmethod
    @add_default_section_for_init("core/task/megatron/supervised")
    def from_core_configure(cls, config, **kwargs):
        """
        Create a MegatronTask instance from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            A MegatronTask instance.
        """
        try:
            torch.distributed.init_process_group(backend="nccl", init_method="env://")
        except:
            logging.info("PyTorch is not in distributed mode")

        config.set_default_section("core/task/megatron/supervised")

        tensor_model_parallel_size = config.getoption(
            "tensor_model_parallel_size",
            1,
        )
        pipeline_model_parallel_size = config.getoption(
            "pipeline_model_parallel_size",
            1,
        )
        context_parallel_size = config.getoption(
            "context_parallel_size",
            1,
        )
        mpu.initialize_model_parallel(
            tensor_model_parallel_size=tensor_model_parallel_size,
            pipeline_model_parallel_size=pipeline_model_parallel_size,
            context_parallel_size=context_parallel_size,
        )

        model = config.getoption("model", None)
        dataset = config.getoption("dataset", None)

        if model is not None:
            model = init_registered_module(model, config, registered_model)

        if dataset is not None:
            dataset = init_registered_module(dataset, config, registered_dataset)

        local_rank = config.getdefault(
            "core/cli",
            "local_rank",
            get_local_rank(),
        )

        return dict(
            configure=config,
            model=model,
            datasets=dataset,
            local_rank=local_rank,
        )

    @add_default_section_for_function("core/task/megatron/supervised")
    def train(
        self,
        optim: str,
        loss_fn: str,
        score_fn: str,
        monitor_fns: Optional[Union[str, List[str]]] = None,
        scheduler: Optional[str] = None,
        from_ckpt_dir: Optional[str] = "./from_ckpt",
        to_ckpt_dir: Optional[str] = "./to_ckpt",
        train_batch_size: Optional[int] = 128,
        dev_batch_size: Optional[int] = 128,
        pin_memory: Optional[bool] = True,
        num_workers: Optional[int] = 4,
        log_freq: Optional[int] = 100,
        ckpt_freq: Optional[int] = 10000,
        grad_acc_step: Optional[int] = 1,
        max_grad_norm: Optional[float] = 1.0,
        num_training_samples: Optional[int] = 1000000000,
        num_dev_samples: Optional[int] = 10000,
        seq_length: Optional[int] = None,
        epochs: Optional[int] = 5,
    ):
        """
        Train the model using deepspeed.

        Args:
            config_path: The path to the deepspeed configuration file.
            optim: The optimizer used for training.
            loss_fn: The loss function used for training.
            score_fn: The score function used for evaluation.
            monitor_fns (optional): The monitoring functions for evaluation. Defaults to None.
            from_ckpt_dir (optional): The directory path to load checkpoints from. Defaults to "./from_ckpt".
            to_ckpt_dir (optional): The directory path to save checkpoints to. Defaults to "./to_ckpt".
            train_batch_size (optional): The batch size for training. Defaults to 128.
            dev_batch_size (optional): The batch size for evaluation. Defaults to 128.
            pin_memory (optional): Whether to pin memory during data loading. Defaults to True.
            num_workers (optional): The number of worker processes for data loading. Defaults to 4.
            save_optimizer (optional): Whether to save the optimizer state. Defaults to False.
            save_scheduler (optional): Whether to save the scheduler state. Defaults to False.
            log_freq (optional): The frequency of logging. Defaults to 100.
            ckpt_freq (optional): The frequency of saving checkpoints. Defaults to 10000.
            grad_acc_step (optional): The number of gradient accumulation steps. Defaults to 1.
            max_grad_norm (optional): The maximum gradient norm. Defaults to 1.0.
            learning_rate (optional): The learning rate for the optimizer. Defaults to None.
            max_warmup_learning_rate (optional): The maximum learning rate during warmup. Defaults to None.
            num_warmup_steps (optional): The number of warmup steps. Defaults to None.
            epochs (optional): The number of training epochs. Defaults to 5.
        """
        if not os.path.exists(to_ckpt_dir) and self.local_rank in [-1, 0]:
            os.makedirs(to_ckpt_dir, exist_ok=True)

        if loss_fn is not None:
            loss_fn = init_registered_module(loss_fn, self.config, registered_loss)

        if score_fn is not None:
            score_fn = init_registered_module(score_fn, self.config, registered_score)

        if monitor_fns is not None:
            monitor_fns = [
                init_registered_module(monitor_fn, self.config, registered_score)
                for monitor_fn in monitor_fns
                if monitor_fn in registered_score
            ]

        if optim is not None and self.model is not None:
            optim = init_registered_module(
                optim,
                self.config,
                registered_optim,
                params=filter(lambda x: x.requires_grad, self.model.parameters()),
            )

        if os.path.exists(from_ckpt_dir):
            self.model.from_checkpoint(from_ckpt_dir)

        if os.path.exists(os.path.join(to_ckpt_dir, "pytorch_model_latest")):
            self.model.from_checkpoint(
                os.path.join(to_ckpt_dir, "pytorch_model_latest")
            )

        info_path = os.path.join(to_ckpt_dir, "info.json")
        if os.path.exists(info_path):
            info = json.load(open(os.path.join(to_ckpt_dir, "info.json")))
        else:
            info = dict()

        global_epoch = info.get("global_epoch", 0)
        global_step = info.get("global_step", 0)
        self.best_score = info.get("best_score", self.best_score)

        logging.info(f"the best score is {self.best_score}")

        for n, p in self.model.named_parameters():
            logging.debug(
                f"{n}: trainable - {p.requires_grad} | tensor dtype - {p.dtype} | tensor shape - {p.shape} | tensor device - {p.device}"
            )

        if self.dp_size > 1:
            self.model = DistributedDataParallel(
                config=self.model.config,
                module=self.model,
                ddp_config=DistributedDataParallelConfig(
                    use_distributed_optimizer=False,
                ),
            )

        dataset_train = self.datasets.get("train")
        dataset_dev = self.datasets.get("dev")

        iter_train = DataLoader(
            dataset_train,
            sampler=None,
            batch_size=train_batch_size,
            shuffle=False,
            pin_memory=pin_memory,
            num_workers=num_workers,
            collate_fn=collate_fn,
        )

        iter_dev = DataLoader(
            dataset_dev,
            sampler=None,
            batch_size=dev_batch_size,
            shuffle=False,
            pin_memory=pin_memory,
            num_workers=num_workers,
            collate_fn=collate_fn,
        )

        iter_train = iter(iter_train)
        iter_dev = iter(iter_dev)

        if scheduler is not None:
            num_training_steps = int(
                epochs
                * num_training_samples
                // train_batch_size
                // max(1, self.dp_size)
                // grad_acc_step
            )

            scheduler = init_registered_module(
                scheduler,
                self.config,
                registered_scheduler,
                optimizer=optim,
                num_training_steps=num_training_steps,
            )

        log_loss = 0
        dev_epoch = 0
        num_train_steps = num_training_samples // train_batch_size // self.dp_size
        num_dev_steps = num_dev_samples // dev_batch_size // self.dp_size

        if self.pp_size == 1:
            forward_backward_pipeline = forward_backward_no_pipelining
        elif self.pp_size > 1 and self.vp_size is not None:
            forward_backward_pipeline = forward_backward_pipelining_with_interleaving
        else:
            forward_backward_pipeline = forward_backward_pipelining_without_interleaving

        def forward_step_func(_iter, _model):
            inputs, targets = next(_iter)
            inputs, targets = inputs.cuda(), targets.cuda()
            inputs = process_batch_data(inputs.dict(), targets.dict())

            outputs = _model(**inputs)
            if isinstance(outputs, LossOutputs):
                loss = outputs.loss
                if self.cp_size > 1:
                    dist.all_reduce(
                        loss,
                        group=self.cp_group,
                    )
                    loss = loss / self.cp_size
                return loss.unsqueeze(0), lambda x: (
                    x.squeeze(0),
                    x.detach().squeeze(0),
                )
            return outputs, None

        for e in range(0, epochs):
            torch.cuda.empty_cache()
            if e < global_epoch:
                continue

            if hasattr(dataset_train, "set_epoch"):
                dataset_train.set_epoch(e)

            if hasattr(dataset_train, "set_skip_step"):
                dataset_train.set_skip_step(global_step * train_batch_size)

            self.model.train()
            for step in range(num_train_steps):
                step = step + global_step

                loss = forward_backward_pipeline(
                    forward_step_func=forward_step_func,
                    data_iterator=iter_train,
                    model=self.model,
                    num_microbatches=grad_acc_step,
                    seq_length=seq_length,
                    micro_batch_size=train_batch_size,
                )

                for param in self.model.parameters():
                    if hasattr(param, "main_grad") and param.main_grad is not None:
                        param.grad = param.main_grad

                if self.is_pp_last_rank:
                    log_loss += loss[0].item()

                nn.utils.clip_grad_norm_(
                    self.model.parameters(),
                    max_norm=max_grad_norm,
                )

                optim.step()
                if scheduler is not None:
                    scheduler.step()
                optim.zero_grad()

                if (
                    (step + 1) % log_freq == 0
                    and self.dp_rank == 0
                    and self.is_pp_last_rank
                    and self.cp_rank == 0
                    and self.tp_rank == 0
                ):
                    logging.info(
                        f"epoch {e} step {step}: loss -- { log_loss / log_freq }"
                    )
                    if wandb.is_available():
                        wandb.log(
                            {
                                "epoch": e,
                                "step": step,
                                "train/loss": log_loss / log_freq,
                            },
                        )
                    log_loss = 0

                if (step + 1) % ckpt_freq == 0:
                    dist.barrier()

                    if hasattr(dataset_dev, "set_epoch"):
                        dataset_dev.set_epoch(dev_epoch)

                    self.model.eval()
                    _log_loss = 0
                    for _ in range(num_dev_steps):
                        _loss = forward_backward_pipeline(
                            forward_step_func=forward_step_func,
                            data_iterator=iter_dev,
                            model=self.model,
                            num_microbatches=grad_acc_step,
                            seq_length=seq_length,
                            micro_batch_size=dev_batch_size,
                            forward_only=True,
                        )

                        if self.is_pp_last_rank:
                            _log_loss += _loss[0].item() / num_dev_steps

                    self.model.train()
                    dev_epoch += 1

                    if (
                        self.dp_rank == 0
                        and self.is_pp_last_rank
                        and self.cp_rank == 0
                        and self.tp_rank == 0
                    ):
                        logging.info(f"loss is { _log_loss }")
                        if wandb.is_available():
                            wandb.log(
                                {
                                    "val/loss": _log_loss,
                                },
                            )

                    if -_log_loss > self.best_score:
                        getattr(self.model, "module", self.model).save_checkpoint(
                            ckpt_dir=os.path.join(to_ckpt_dir, "pytorch_model"),
                            weight_name="common.pt",
                        )
                        self.best_score = -_log_loss
                        info["best_score"] = self.best_score
                        info["global_epoch"] = e
                        info["global_step"] = step + 1
                        with open(info_path, "w") as f:
                            json.dump(info, f, indent=4)

                    getattr(self.model, "module", self.model).save_checkpoint(
                        ckpt_dir=os.path.join(to_ckpt_dir, "pytorch_model_latest"),
                        weight_name="common.pt",
                    )
                    info["best_score"] = self.best_score
                    info["global_epoch"] = e
                    info["global_step"] = step + 1
                    with open(info_path, "w") as f:
                        json.dump(info, f, indent=4)

                    dist.barrier()

            log_loss = 0

            dist.barrier()

            if hasattr(dataset_dev, "set_epoch"):
                dataset_dev.set_epoch(dev_epoch)

            self.model.eval()
            _log_loss = 0
            for _ in range(num_dev_steps):
                _loss = forward_backward_pipeline(
                    forward_step_func=forward_step_func,
                    data_iterator=iter_dev,
                    model=self.model,
                    num_microbatches=grad_acc_step,
                    seq_length=seq_length,
                    micro_batch_size=dev_batch_size,
                    forward_only=True,
                )

                if self.is_pp_last_rank:
                    _log_loss += _loss[0].item() / num_dev_steps

            self.model.train()
            dev_epoch += 1

            if (
                self.dp_rank == 0
                and self.is_pp_last_rank
                and self.cp_rank == 0
                and self.tp_rank == 0
            ):
                logging.info(f"loss is { _log_loss }")
                if wandb.is_available():
                    wandb.log(
                        {
                            "val/loss": _log_loss,
                        },
                    )

            if -_log_loss > self.best_score:
                getattr(self.model, "module", self.model).save_checkpoint(
                    ckpt_dir=os.path.join(to_ckpt_dir, "pytorch_model"),
                    weight_name="common.pt",
                )
                self.best_score = -_log_loss
                info["best_score"] = self.best_score
                info["global_epoch"] = e
                info["global_step"] = step + 1
                with open(info_path, "w") as f:
                    json.dump(info, f, indent=4)

            getattr(self.model, "module", self.model).save_checkpoint(
                ckpt_dir=os.path.join(to_ckpt_dir, "pytorch_model_latest"),
                weight_name="common.pt",
            )
            info["best_score"] = self.best_score
            info["global_epoch"] = e
            info["global_step"] = step + 1
            with open(info_path, "w") as f:
                json.dump(info, f, indent=4)

            dist.barrier()

            global_step = 0



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union


def sort_by_scores(
    scores,
    features_list,
    topn=None,
    shuffle_ties=False,
):
    """Sorts example features according to per-example scores.
    Args:
      scores: A `Tensor` of shape [batch_size, list_size] representing the
        per-example scores.
      features_list: A list of `Tensor`s with the same shape as scores to be
        sorted.
      topn: An integer as the cutoff of examples in the sorted list.
      shuffle_ties: A boolean. If True, randomly shuffle before the sorting.
    Returns:
      A list of `Tensor`s as the list of sorted features by `scores`.
    """
    score_size = list(scores.size())
    assert len(score_size) == 2
    for f in features_list:
        assert len(f.size()) == 2
    list_size = score_size[1]
    if topn is None:
        topn = list_size
    topn = min(topn, list_size)
    if shuffle_ties:
        shuffle_ind = torch.argsort(torch.rand(score_size)).to(scores.device)
        scores = torch.gather(scores, dim=1, index=shuffle_ind)
        features_list = [
            torch.gather(f, dim=1, index=shuffle_ind) for f in features_list
        ]
    _, indices = scores.topk(topn, sorted=True)
    return [torch.gather(f, dim=1, index=indices) for f in features_list]


def approx_ranks(logits, alpha=10.0):
    r"""Computes approximate ranks given a list of logits.

    Given a list of logits, the rank of an item in the list is simply
    one plus the total number of items with a larger logit. In other words,

      rank_i = 1 + \sum_{j \neq i} I_{s_j > s_i},

    where "I" is the indicator function. The indicator function can be
    approximated by a generalized sigmoid:

      I_{s_j < s_i} \approx 1/(1 + exp(-\alpha * (s_j - s_i))).

    This function approximates the rank of an item using this sigmoid
    approximation to the indicator function. This technique is at the core
    of "A general approximation framework for direct optimization of
    information retrieval measures" by Qin et al.

    Args:
      logits: A `Tensor` with shape [batch_size, list_size]. Each value is the
        ranking score of the corresponding item.
      alpha: Exponent of the generalized sigmoid function.

    Returns:
      A `Tensor` of ranks with the same shape as logits.
    """
    """
    list_size=tf.shape(input=logits)[1]
    x = tf.tile(tf.expand_dims(logits, 2), [1, 1, list_size])
    y = tf.tile(tf.expand_dims(logits, 1), [1, list_size, 1])
    pairs = tf.sigmoid(alpha * (y - x))
    return tf.reduce_sum(input_tensor=pairs, axis=-1) + .5
    """
    x = torch.unsqueeze(logits, 2)
    y = torch.unsqueeze(logits, 1)
    pairs = torch.sigmoid(alpha * (y - x))
    return torch.sum(pairs, -1) + 0.5


def inverse_max_dcg(
    labels,
    gain_fn=lambda labels: torch.pow(2.0, labels) - 1.0,
    rank_discount_fn=lambda rank: 1.0 / torch.log1p(rank),
    topn=None,
):
    """Computes the inverse of max DCG.

    Args:
      labels: A `Tensor` with shape [batch_size, list_size]. Each value is the
        graded relevance of the corresponding item.
      gain_fn: A gain function. By default this is set to: 2^label - 1.
      rank_discount_fn: A discount function. By default this is set to:
        1/log(1+rank).
      topn: An integer as the cutoff of examples in the sorted list.

    Returns:
      A `Tensor` with shape [batch_size, 1].
    """
    (ideal_sorted_labels,) = sort_by_scores(labels, [labels], topn=topn)
    rank = torch.arange(ideal_sorted_labels.size(1)) + 1
    discounted_gain = gain_fn(ideal_sorted_labels) * rank_discount_fn(
        rank.float().to(ideal_sorted_labels.device)
    )
    discounted_gain = torch.sum(discounted_gain, 1, keepdim=True)
    return torch.where(
        torch.gt(discounted_gain, 0.0),
        1.0 / discounted_gain,
        torch.zeros_like(discounted_gain),
    )


def ndcg(
    labels,
    ranks=None,
    perm_mat=None,
    gain_fn=lambda labels: torch.pow(2.0, labels) - 1.0,
    rank_discount_fn=lambda rank: 1.0 / torch.log1p(rank),
):
    """Computes NDCG from labels and ranks.

    Args:
      labels: A `Tensor` with shape [batch_size, list_size], representing graded
        relevance.
      ranks: A `Tensor` of the same shape as labels, or [1, list_size], or None.
        If ranks=None, we assume the labels are sorted in their rank.
      perm_mat: A `Tensor` with shape [batch_size, list_size, list_size] or None.
        Permutation matrices with rows correpond to the ranks and columns
        correspond to the indices. An argmax over each row gives the index of the
        element at the corresponding rank.

    Returns:
      A `tensor` of NDCG, ApproxNDCG, or ExpectedNDCG of shape [batch_size, 1].
    """
    if ranks is not None and perm_mat is not None:
        raise ValueError("Cannot use both ranks and perm_mat simultaneously.")

    if ranks is None:
        list_size = labels.size(1)
        ranks = torch.arange(list_size) + 1
    discounts = rank_discount_fn(ranks.float())
    gains = gain_fn(labels.float())
    if perm_mat is not None:
        gains = torch.sum(perm_mat * torch.unsqueeze(gains, 1), -1)
    dcg = torch.sum(gains * discounts, -1, keepdim=True)
    return dcg * inverse_max_dcg(
        labels,
        gain_fn=gain_fn,
        rank_discount_fn=rank_discount_fn,
    )


class ListMLELoss(nn.Module):
    def __init__(
        self,
        reduction: Optional[str] = "mean",
    ):
        super().__init__()
        self.reduction = reduction

    def forward(
        self,
        input: torch.Tensor,
        target: torch.Tensor,
        masks: torch.Tensor,
        sample_weight: Optional[torch.Tensor] = None,
    ):
        target = target * masks
        input = input + (1 - masks) * torch.tensor(1e-6).log()
        scores = torch.where(
            masks.bool(),
            target,
            torch.min(target, axis=1, keepdims=True)[0] - 1e-6,
        )
        sorted_labels, sorted_logits = sort_by_scores(
            scores,
            [target, input],
            shuffle_ties=True,
        )
        raw_max = torch.max(sorted_logits, dim=1, keepdim=True)[0]
        sorted_logits = sorted_logits - raw_max

        sums = torch.flip(
            torch.cumsum(torch.flip(torch.exp(sorted_logits), [-1]), dim=1),
            [-1],
        )

        sums = torch.log(sums) - sorted_logits
        loss = torch.sum(sums, dim=1)

        if sample_weight is not None:
            loss = loss * sample_weight

        if self.reduction == "mean":
            loss = torch.mean(loss)

        return loss


class ApproxNDCGLoss(nn.Module):
    def __init__(
        self,
        reduction: Optional[str] = "mean",
        alpha: Optional[float] = 10.0,
    ):
        super().__init__()
        self.reduction = reduction
        self.alpha = alpha

    def forward(
        self,
        input: torch.Tensor,
        target: torch.Tensor,
        masks: torch.Tensor,
        sample_weight: Optional[torch.Tensor] = None,
    ):
        target = target * masks
        input = input + (1 - masks) * (torch.min(input, -1, keepdim=True)[0] - 1e3)
        ranks = approx_ranks(input, alpha=self.alpha)
        loss = -ndcg(target, ranks)

        if sample_weight is not None:
            loss = loss * sample_weight

        if self.reduction == "mean":
            loss = torch.mean(loss)

        return loss


class ApproxMRRLoss(nn.Module):
    def __init__(
        self,
        reduction: Optional[str] = "mean",
        alpha: Optional[float] = 10.0,
    ):
        super().__init__()
        self.reduction = reduction
        self.alpha = alpha

    def forward(
        self,
        input: torch.Tensor,
        target: torch.Tensor,
        masks: torch.Tensor,
        sample_weight: Optional[torch.Tensor] = None,
    ):
        target = target * masks
        input = input + (1 - masks) * (torch.min(input, -1, keepdim=True)[0] - 1e3)
        rr = 1.0 / approx_ranks(input, alpha=self.alpha)
        rr = torch.sum(input * target, -1, keepdim=True)
        mrr = rr / torch.sum(target, -1, keepdim=True)
        loss = -mrr

        if sample_weight is not None:
            loss = loss * sample_weight

        if self.reduction == "mean":
            loss = torch.mean(loss)

        return loss



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
from transformers.activations import quick_gelu
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union


class reslayer(nn.Module):
    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        output_dim: int,
        use_bn: Optional[bool] = True,
    ):
        super().__init__()
        assert input_dim == output_dim
        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, output_dim)
        self.bn = nn.BatchNorm1d(hidden_dim) if use_bn else None

        # init weight
        self.fc1.weight.data.normal_(mean=0.0, std=0.02)
        self.fc1.bias.data.zero_()

        self.fc2.weight.data.normal_(mean=0.0, std=0.02)
        self.fc2.bias.data.zero_()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        output = self.fc1(input)
        if self.bn is not None:
            output = self.bn(output)
        output = torch.relu(output)
        output = self.fc2(output) + input
        return torch.relu(output)


class mlplayer(nn.Module):
    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        output_dim: int,
        add_pre_layer_norm: Optional[bool] = True,
        add_post_layer_norm: Optional[bool] = False,
    ):
        super().__init__()
        assert input_dim == output_dim
        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, output_dim)

        # Initialize weights
        self.fc1.weight.data.normal_(mean=0.0, std=0.02)
        self.fc1.bias.data.zero_()

        self.fc2.weight.data.normal_(mean=0.0, std=0.02)
        self.fc2.bias.data.zero_()

        self.add_pre_layer_norm = add_pre_layer_norm
        self.pre_layer_norm = nn.LayerNorm(output_dim) if add_pre_layer_norm else None

        self.add_post_layer_norm = add_post_layer_norm
        self.post_layer_norm = nn.LayerNorm(output_dim) if add_post_layer_norm else None

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.add_pre_layer_norm:
            output = self.pre_layer_norm(input)
        else:
            output = input
        output = self.fc1(output)
        output = quick_gelu(output)
        output = self.fc2(output) + input
        if self.add_post_layer_norm:
            output = self.post_layer_norm(output)
        return output



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
import torch
from torch import autocast
import torch.nn.functional as F
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
import diffusers.schedulers as schedulers
from transformers import CLIPTextConfig, CLIPTextModel, CLIPTextModelWithProjection
from transformers.models.t5.configuration_t5 import T5Config
from transformers.models.t5.modeling_t5 import T5EncoderModel
from diffusers.schedulers import SchedulerMixin
from diffusers.training_utils import (
    compute_loss_weighting_for_sd3,
    compute_density_for_timestep_sampling,
)
from diffusers.pipelines import (
    FluxPipeline,
    FluxControlNetPipeline,
    FluxControlNetImg2ImgPipeline,
    FluxControlNetInpaintPipeline,
)
from unitorch.models import (
    GenericModel,
    GenericOutputs,
    QuantizationConfig,
    QuantizationMixin,
)
from unitorch.models.diffusers import GenericStableFluxModel
from unitorch.models.diffusers.modeling_stable_flux import (
    _prepare_latent_image_ids,
    _pack_latents,
    _unpack_latents,
)
from unitorch.models.diffusers import compute_snr


class ControlNetFluxForText2ImageGeneration(GenericStableFluxModel):
    """
    ControlNetXL model for text-to-image generation.

    Args:
        config_path (str): Path to the model configuration file.
        text_config_path (str): Path to the text model configuration file.
        text2_config_path (str): Path to the second text model configuration file.
        vae_config_path (str): Path to the VAE model configuration file.
        controlnet_configs_path (str): Path to the ControlNet model configuration file.
        scheduler_config_path (str): Path to the scheduler configuration file.
        quant_config_path (Optional[str]): Path to the quantization configuration file (default: None).
        image_size (Optional[int]): Size of the input image (default: None).
        in_channels (Optional[int]): Number of input channels (default: None).
        out_channels (Optional[int]): Number of output channels (default: None).
        num_train_timesteps (Optional[int]): Number of training timesteps (default: 1000).
        num_infer_timesteps (Optional[int]): Number of inference timesteps (default: 50).
        freeze_vae_encoder (Optional[bool]): Whether to freeze the VAE encoder (default: True).
        freeze_text_encoder (Optional[bool]): Whether to freeze the text encoder (default: True).
        freeze_transformer_encoder (Optional[bool]): Whether to freeze the transformer encoder (default: True).
        seed (Optional[int]): Random seed (default: 1123).
    """

    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        text2_config_path: str,
        vae_config_path: str,
        controlnet_configs_path: Union[str, List[str]],
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        freeze_transformer_encoder: Optional[bool] = False,
        snr_gamma: Optional[float] = 5.0,
        seed: Optional[int] = 1123,
        guidance_scale: Optional[float] = 3.5,
        controlnet_conditioning_mode: Optional[Union[int, List[int]]] = None,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            text2_config_path=text2_config_path,
            vae_config_path=vae_config_path,
            controlnet_configs_path=controlnet_configs_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            freeze_transformer_encoder=freeze_transformer_encoder,
            snr_gamma=snr_gamma,
            seed=seed,
        )
        self.pipeline = FluxControlNetPipeline(
            vae=self.vae,
            text_encoder=self.text,
            text_encoder_2=self.text2,
            transformer=self.transformer,
            controlnet=self.controlnet,
            scheduler=self.scheduler,
            tokenizer=None,
            tokenizer_2=None,
        )
        self.guidance_scale = guidance_scale
        self.pipeline.set_progress_bar_config(disable=True)
        self.controlnet_conditioning_mode = controlnet_conditioning_mode

    def forward(
        self,
        condition_pixel_values: torch.Tensor,
        input_ids: torch.Tensor,
        input2_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        attention2_mask: Optional[torch.Tensor] = None,
    ):
        outputs = self.get_prompt_outputs(
            input_ids=input_ids,
            input2_ids=input2_ids,
            attention_mask=attention_mask,
            attention2_mask=attention2_mask,
        )
        latents = self.vae.encode(pixel_values).latent_dist.sample()
        latents = (
            latents - self.vae.config.shift_factor
        ) * self.vae.config.scaling_factor
        latent_image_ids = _prepare_latent_image_ids(
            latents.shape[0],
            latents.shape[2] // 2,
            latents.shape[3] // 2,
            self.device,
            self.dtype,
        )

        noise = torch.randn(latents.shape).to(latents.device)
        batch = latents.shape[0]

        u = compute_density_for_timestep_sampling(
            weighting_scheme="none",
            batch_size=batch,
            logit_mean=0.0,
            logit_std=1.0,
            mode_scale=1.29,
        )
        indices = (u * self.scheduler.config.num_train_timesteps).long()
        timesteps = self.scheduler.timesteps[indices].to(device=self.device)

        sigmas = self.get_sigmas(timesteps, n_dim=latents.ndim, dtype=latents.dtype)
        noise_latents = (1.0 - sigmas) * latents + sigmas * noise

        noise_latents = _pack_latents(
            noise_latents,
            batch_size=latents.shape[0],
            num_channels_latents=latents.shape[1],
            height=latents.shape[2],
            width=latents.shape[3],
        )

        text_ids = torch.zeros(outputs.prompt_embeds.shape[1], 3).to(
            device=self.device, dtype=self.dtype
        )

        if self.num_controlnets == 1:
            condition_latents = self.vae.encode(
                condition_pixel_values
            ).latent_dist.sample()
            condition_latents = (
                condition_latents - self.vae.config.shift_factor
            ) * self.vae.config.scaling_factor
            condition_latents = _pack_latents(
                condition_latents,
                batch_size=condition_latents.shape[0],
                num_channels_latents=condition_latents.shape[1],
                height=condition_latents.shape[2],
                width=condition_latents.shape[3],
            )
        else:
            condition_pixel_values = condition_pixel_values.transpose(0, 1)
            condition_latents = [
                self.vae.encode(_condition_pixel_values).latent_dist.sample()
                for _condition_pixel_values in list(condition_pixel_values)
            ]
            condition_latents = [
                (latent - self.vae.config.shift_factor) * self.vae.config.scaling_factor
                for latent in condition_latents
            ]
            condition_latents = [
                _pack_latents(
                    latent,
                    batch_size=latent.shape[0],
                    num_channels_latents=latent.shape[1],
                    height=latent.shape[2],
                    width=latent.shape[3],
                )
                for latent in condition_latents
            ]

        if self.num_controlnets == 1:
            use_guidance = self.controlnet.config.guidance_embeds
        else:
            use_guidance = [net.config.guidance_embeds for net in self.controlnet.nets]
            assert all(g == use_guidance[0] for g in use_guidance)
            use_guidance = use_guidance[0]

        if use_guidance:
            guidance = torch.full(
                [1], self.guidance_scale, device=self.device, dtype=torch.float32
            )
            guidance = guidance.expand(latents.shape[0])
        else:
            guidance = None

        if self.controlnet_conditioning_mode is None:
            controlnet_mode = (
                None if self.num_controlnets == 1 else [None] * self.num_controlnets
            )
        else:
            if self.num_controlnets == 1:
                controlnet_mode = (
                    torch.tensor(self.controlnet_conditioning_mode)
                    .to(self.device)
                    .long()
                )
                controlnet_mode = controlnet_mode.expand(latents.shape[0]).reshape(
                    -1, 1
                )
            else:
                assert (
                    isinstance(self.controlnet_conditioning_mode, list)
                    and len(self.controlnet_conditioning_mode) == self.num_controlnets
                )

                def _get_mode(mode):
                    if mode is None:
                        return None
                    _mode = torch.tensor(mode).to(self.device).long()
                    return _mode.expand(latents.shape[0]).reshape(-1)

                controlnet_mode = [
                    _get_mode(mode) for mode in self.controlnet_conditioning_mode
                ]
        controlnet_block_samples, controlnet_single_block_samples = self.controlnet(
            hidden_states=noise_latents,
            timestep=timesteps / 1000,
            guidance=guidance,
            encoder_hidden_states=outputs.prompt_embeds,
            pooled_projections=outputs.pooled_prompt_embeds,
            controlnet_cond=condition_latents,
            controlnet_mode=controlnet_mode,
            conditioning_scale=1.0
            if self.num_controlnets == 1
            else [1.0] * self.num_controlnets,
            txt_ids=text_ids,
            img_ids=latent_image_ids,
            return_dict=False,
        )

        if self.transformer.config.guidance_embeds:
            guidance = torch.full(
                [1], self.guidance_scale, device=self.device, dtype=torch.float32
            )
            guidance = guidance.expand(latents.shape[0])
        else:
            guidance = None

        outputs = self.transformer(
            hidden_states=noise_latents,
            timestep=timesteps / 1000,
            guidance=guidance,
            encoder_hidden_states=outputs.prompt_embeds,
            pooled_projections=outputs.pooled_prompt_embeds,
            controlnet_block_samples=controlnet_block_samples,
            controlnet_single_block_samples=controlnet_single_block_samples,
            txt_ids=text_ids,
            img_ids=latent_image_ids,
            return_dict=False,
        )[0]

        vae_scale_factor = 2 ** (len(self.vae.config.block_out_channels) - 1)
        outputs = _unpack_latents(
            outputs,
            height=latents.shape[2] * vae_scale_factor,
            width=latents.shape[3] * vae_scale_factor,
            vae_scale_factor=vae_scale_factor,
        )

        weighting = compute_loss_weighting_for_sd3(
            weighting_scheme="none", sigmas=sigmas
        )
        target = noise - latents
        loss = torch.mean(
            (weighting.float() * (outputs.float() - target.float()) ** 2).reshape(
                target.shape[0], -1
            ),
            1,
        )
        loss = loss.mean()
        return loss

    def generate(
        self,
        condition_pixel_values: torch.Tensor,
        input_ids: torch.Tensor,
        input2_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        attention2_mask: Optional[torch.Tensor] = None,
        height: Optional[int] = 1024,
        width: Optional[int] = 1024,
        guidance_scale: Optional[float] = 5.0,
        controlnet_conditioning_scale: Optional[Union[float, List[float]]] = None,
        controlnet_conditioning_mode: Optional[Union[int, List[int]]] = None,
    ):
        """
        Generate images using the model.
        Args:
            condition_pixel_values (torch.Tensor): Condition pixel values.
            input_ids (torch.Tensor): Input IDs.
            input2_ids (torch.Tensor): Second input IDs.
            negative_input_ids (torch.Tensor): Negative input IDs.
            negative_input2_ids (torch.Tensor): Negative second input IDs.
            attention_mask (Optional[torch.Tensor]): Attention mask (default: None).
            attention2_mask (Optional[torch.Tensor]): Second attention mask (default: None).
            negative_attention_mask (Optional[torch.Tensor]): Negative attention mask (default: None).
            negative_attention2_mask (Optional[torch.Tensor]): Negative second attention mask (default: None).
            height (Optional[int]): Height of the generated images (default: 1024).
            width (Optional[int]): Width of the generated images (default: 1024).
            guidance_scale (Optional[float]): Scale for guidance (default: 5.0).

        Returns:
            GenericOutputs: Generated images.
        """
        controlnet_conditioning_mode = (
            controlnet_conditioning_mode or self.controlnet_conditioning_mode
        )
        outputs = self.get_prompt_outputs(
            input_ids=input_ids,
            input2_ids=input2_ids,
            attention_mask=attention_mask,
            attention2_mask=attention2_mask,
        )
        if controlnet_conditioning_scale is None:
            if self.num_controlnets == 1:
                controlnet_conditioning_scale = 1.0
            else:
                controlnet_conditioning_scale = [1.0] * self.num_controlnets
        elif (
            not isinstance(controlnet_conditioning_scale, list)
            and self.num_controlnets > 1
        ):
            controlnet_conditioning_scale = [
                controlnet_conditioning_scale
            ] * self.num_controlnets

        if controlnet_conditioning_mode is None:
            if self.num_controlnets == 1:
                controlnet_conditioning_mode = None
            else:
                controlnet_conditioning_mode = [None] * self.num_controlnets

        images = self.pipeline(
            control_image=condition_pixel_values
            if self.num_controlnets == 1
            else list(condition_pixel_values.transpose(0, 1)),
            control_mode=controlnet_conditioning_mode,
            prompt_embeds=outputs.prompt_embeds,
            pooled_prompt_embeds=outputs.pooled_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            num_inference_steps=self.num_infer_timesteps,
            height=height,
            width=width,
            guidance_scale=guidance_scale,
            controlnet_conditioning_scale=controlnet_conditioning_scale,
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))


class ControlNetFluxForImage2ImageGeneration(GenericStableFluxModel):
    """
    ControlNetXL model for text-to-image generation.

    Args:
        config_path (str): Path to the model configuration file.
        text_config_path (str): Path to the text model configuration file.
        text2_config_path (str): Path to the second text model configuration file.
        vae_config_path (str): Path to the VAE model configuration file.
        controlnet_configs_path (str): Path to the ControlNet model configuration file.
        scheduler_config_path (str): Path to the scheduler configuration file.
        quant_config_path (Optional[str]): Path to the quantization configuration file (default: None).
        image_size (Optional[int]): Size of the input image (default: None).
        in_channels (Optional[int]): Number of input channels (default: None).
        out_channels (Optional[int]): Number of output channels (default: None).
        num_train_timesteps (Optional[int]): Number of training timesteps (default: 1000).
        num_infer_timesteps (Optional[int]): Number of inference timesteps (default: 50).
        freeze_vae_encoder (Optional[bool]): Whether to freeze the VAE encoder (default: True).
        freeze_text_encoder (Optional[bool]): Whether to freeze the text encoder (default: True).
        freeze_transformer_encoder (Optional[bool]): Whether to freeze the transformer encoder (default: True).
        seed (Optional[int]): Random seed (default: 1123).
    """

    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        text2_config_path: str,
        vae_config_path: str,
        controlnet_configs_path: Union[str, List[str]],
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        freeze_transformer_encoder: Optional[bool] = False,
        snr_gamma: Optional[float] = 5.0,
        seed: Optional[int] = 1123,
        controlnet_conditioning_mode: Optional[Union[int, List[int]]] = None,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            text2_config_path=text2_config_path,
            vae_config_path=vae_config_path,
            controlnet_configs_path=controlnet_configs_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            freeze_transformer_encoder=freeze_transformer_encoder,
            snr_gamma=snr_gamma,
            seed=seed,
        )
        self.pipeline = FluxControlNetImg2ImgPipeline(
            vae=self.vae,
            text_encoder=self.text,
            text_encoder_2=self.text2,
            transformer=self.transformer,
            controlnet=self.controlnet,
            scheduler=self.scheduler,
            tokenizer=None,
            tokenizer_2=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)
        self.controlnet_conditioning_mode = controlnet_conditioning_mode

    def forward(
        self,
    ):
        raise NotImplementedError

    def generate(
        self,
        pixel_values: torch.Tensor,
        condition_pixel_values: torch.Tensor,
        input_ids: torch.Tensor,
        input2_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        attention2_mask: Optional[torch.Tensor] = None,
        strength: Optional[float] = 0.99,
        guidance_scale: Optional[float] = 5.0,
        controlnet_conditioning_scale: Optional[Union[float, List[float]]] = None,
        controlnet_conditioning_mode: Optional[Union[int, List[int]]] = None,
    ):
        """
        Generate images using the model.

        Args:
            condition_pixel_values (torch.Tensor): Condition pixel values.
            input_ids (torch.Tensor): Input IDs.
            input2_ids (torch.Tensor): Second input IDs.
            negative_input_ids (torch.Tensor): Negative input IDs.
            negative_input2_ids (torch.Tensor): Negative second input IDs.
            attention_mask (Optional[torch.Tensor]): Attention mask (default: None).
            attention2_mask (Optional[torch.Tensor]): Second attention mask (default: None).
            negative_attention_mask (Optional[torch.Tensor]): Negative attention mask (default: None).
            negative_attention2_mask (Optional[torch.Tensor]): Negative second attention mask (default: None).
            height (Optional[int]): Height of the generated images (default: 1024).
            width (Optional[int]): Width of the generated images (default: 1024).
            guidance_scale (Optional[float]): Scale for guidance (default: 5.0).

        Returns:
            GenericOutputs: Generated images.
        """
        controlnet_conditioning_mode = (
            controlnet_conditioning_mode or self.controlnet_conditioning_mode
        )
        outputs = self.get_prompt_outputs(
            input_ids=input_ids,
            input2_ids=input2_ids,
            attention_mask=attention_mask,
            attention2_mask=attention2_mask,
        )
        if controlnet_conditioning_scale is None:
            if self.num_controlnets == 1:
                controlnet_conditioning_scale = 1.0
            else:
                controlnet_conditioning_scale = [1.0] * self.num_controlnets
        elif (
            not isinstance(controlnet_conditioning_scale, list)
            and self.num_controlnets > 1
        ):
            controlnet_conditioning_scale = [
                controlnet_conditioning_scale
            ] * self.num_controlnets

        if controlnet_conditioning_mode is None:
            if self.num_controlnets == 1:
                controlnet_conditioning_mode = None
            else:
                controlnet_conditioning_mode = [None] * self.num_controlnets

        images = self.pipeline(
            image=pixel_values,
            control_image=condition_pixel_values
            if self.num_controlnets == 1
            else list(condition_pixel_values.transpose(0, 1)),
            control_mode=controlnet_conditioning_mode,
            prompt_embeds=outputs.prompt_embeds,
            pooled_prompt_embeds=outputs.pooled_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            num_inference_steps=self.num_infer_timesteps,
            strength=strength,
            guidance_scale=guidance_scale,
            controlnet_conditioning_scale=controlnet_conditioning_scale,
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))


class ControlNetFluxForImageInpainting(GenericStableFluxModel):
    """
    ControlNetXL model for text-to-image generation.

    Args:
        config_path (str): Path to the model configuration file.
        text_config_path (str): Path to the text model configuration file.
        text2_config_path (str): Path to the second text model configuration file.
        vae_config_path (str): Path to the VAE model configuration file.
        controlnet_configs_path (str): Path to the ControlNet model configuration file.
        scheduler_config_path (str): Path to the scheduler configuration file.
        quant_config_path (Optional[str]): Path to the quantization configuration file (default: None).
        image_size (Optional[int]): Size of the input image (default: None).
        in_channels (Optional[int]): Number of input channels (default: None).
        out_channels (Optional[int]): Number of output channels (default: None).
        num_train_timesteps (Optional[int]): Number of training timesteps (default: 1000).
        num_infer_timesteps (Optional[int]): Number of inference timesteps (default: 50).
        freeze_vae_encoder (Optional[bool]): Whether to freeze the VAE encoder (default: True).
        freeze_text_encoder (Optional[bool]): Whether to freeze the text encoder (default: True).
        freeze_transformer_encoder (Optional[bool]): Whether to freeze the transformer encoder (default: True).
        seed (Optional[int]): Random seed (default: 1123).
    """

    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        text2_config_path: str,
        vae_config_path: str,
        controlnet_configs_path: Union[str, List[str]],
        scheduler_config_path: str,
        inpainting_controlnet_config_path: Union[str] = None,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        freeze_transformer_encoder: Optional[bool] = False,
        snr_gamma: Optional[float] = 5.0,
        seed: Optional[int] = 1123,
        controlnet_conditioning_mode: Optional[Union[int, List[int]]] = None,
        inpainting_controlnet_conditioning_mode: Optional[Union[int, List[int]]] = None,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            text2_config_path=text2_config_path,
            vae_config_path=vae_config_path,
            controlnet_configs_path=controlnet_configs_path,
            inpainting_controlnet_config_path=inpainting_controlnet_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            freeze_transformer_encoder=freeze_transformer_encoder,
            snr_gamma=snr_gamma,
            seed=seed,
        )
        self.pipeline = FluxControlNetInpaintPipeline(
            vae=self.vae,
            text_encoder=self.text,
            text_encoder_2=self.text2,
            transformer=self.transformer,
            controlnet=self.controlnet,
            scheduler=self.scheduler,
            tokenizer=None,
            tokenizer_2=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)
        self.controlnet_conditioning_mode = controlnet_conditioning_mode
        self.inpainting_controlnet_conditioning_mode = (
            inpainting_controlnet_conditioning_mode
        )

    def forward(
        self,
    ):
        raise NotImplementedError

    def generate(
        self,
        pixel_values: torch.Tensor,
        pixel_masks: torch.Tensor,
        input_ids: torch.Tensor,
        input2_ids: torch.Tensor,
        condition_pixel_values: torch.Tensor = None,
        inpainting_condition_pixel_values: torch.Tensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        attention2_mask: Optional[torch.Tensor] = None,
        strength: Optional[float] = 1.0,
        guidance_scale: Optional[float] = 5.0,
        controlnet_conditioning_scale: Optional[Union[float, List[float]]] = None,
        controlnet_conditioning_mode: Optional[Union[int, List[int]]] = None,
        inpainting_controlnet_conditioning_scale: Optional[float] = None,
        inpainting_controlnet_conditioning_mode: Optional[int] = None,
    ):
        """
        Generate images using the model.

        Args:
            condition_pixel_values (torch.Tensor): Condition pixel values.
            input_ids (torch.Tensor): Input IDs.
            input2_ids (torch.Tensor): Second input IDs.
            negative_input_ids (torch.Tensor): Negative input IDs.
            negative_input2_ids (torch.Tensor): Negative second input IDs.
            attention_mask (Optional[torch.Tensor]): Attention mask (default: None).
            attention2_mask (Optional[torch.Tensor]): Second attention mask (default: None).
            negative_attention_mask (Optional[torch.Tensor]): Negative attention mask (default: None).
            negative_attention2_mask (Optional[torch.Tensor]): Negative second attention mask (default: None).
            height (Optional[int]): Height of the generated images (default: 1024).
            width (Optional[int]): Width of the generated images (default: 1024).
            guidance_scale (Optional[float]): Scale for guidance (default: 5.0).

        Returns:
            GenericOutputs: Generated images.
        """
        assert (
            condition_pixel_values is not None
            or inpainting_condition_pixel_values is not None
        )
        controlnet_conditioning_mode = (
            controlnet_conditioning_mode or self.controlnet_conditioning_mode
        )
        inpainting_controlnet_conditioning_mode = (
            inpainting_controlnet_conditioning_mode
            or self.inpainting_controlnet_conditioning_mode
        )
        if inpainting_condition_pixel_values is not None:
            if self.num_controlnets == 1:
                condition_pixel_values = inpainting_condition_pixel_values
                controlnet_conditioning_scale = inpainting_controlnet_conditioning_scale
                controlnet_conditioning_mode = inpainting_controlnet_conditioning_mode
            else:
                condition_pixel_values = torch.cat(
                    [
                        condition_pixel_values,
                        inpainting_condition_pixel_values.unsqueeze(1),
                    ],
                    dim=1,
                )
                if controlnet_conditioning_scale is None:
                    controlnet_conditioning_scale = [1.0] * (self.num_controlnets - 1)
                controlnet_conditioning_scale += [
                    inpainting_controlnet_conditioning_scale
                ]
                if controlnet_conditioning_mode is None:
                    controlnet_conditioning_mode = [None] * (self.num_controlnets - 1)
                controlnet_conditioning_mode += [
                    inpainting_controlnet_conditioning_mode
                ]
        else:
            if controlnet_conditioning_scale is None:
                if self.num_controlnets == 1:
                    controlnet_conditioning_scale = 1.0
                else:
                    controlnet_conditioning_scale = [1.0] * self.num_controlnets
            elif (
                not isinstance(controlnet_conditioning_scale, list)
                and self.num_controlnets > 1
            ):
                controlnet_conditioning_scale = [
                    controlnet_conditioning_scale
                ] * self.num_controlnets

            if controlnet_conditioning_mode is None:
                if self.num_controlnets == 1:
                    controlnet_conditioning_mode = None
                else:
                    controlnet_conditioning_mode = [None] * self.num_controlnets

        outputs = self.get_prompt_outputs(
            input_ids=input_ids,
            input2_ids=input2_ids,
            attention_mask=attention_mask,
            attention2_mask=attention2_mask,
        )

        images = self.pipeline(
            image=pixel_values,
            mask_image=pixel_masks,
            control_image=condition_pixel_values
            if self.num_controlnets == 1
            else list(condition_pixel_values.transpose(0, 1)),
            control_mode=controlnet_conditioning_mode,
            prompt_embeds=outputs.prompt_embeds,
            pooled_prompt_embeds=outputs.pooled_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            num_inference_steps=self.num_infer_timesteps,
            width=pixel_values.size(-1),
            height=pixel_values.size(-2),
            strength=strength,
            guidance_scale=guidance_scale,
            controlnet_conditioning_scale=controlnet_conditioning_scale,
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.siglip.modeling import (
    SiglipForPretrain,
    SiglipForClassification,
    SiglipForTextClassification,
    SiglipForImageClassification,
    SiglipForMatching,
)
from unitorch.models.siglip.processing import SiglipProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.fastapis.stable_xl.text2image import (
    StableXLForText2ImageFastAPIPipeline,
)
from unitorch.cli.fastapis.stable_xl.image2image import (
    StableXLForImage2ImageFastAPIPipeline,
)
from unitorch.cli.fastapis.stable_xl.inpainting import (
    StableXLForImageInpaintingFastAPIPipeline,
)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from functools import partial

import torch
from transformers import XLMRobertaTokenizer
from unitorch.utils import pop_value, truncate_sequence_pair
from unitorch.models import GenericOutputs, HfTextClassificationProcessor


def get_xlm_roberta_tokenizer(vocab_path):
    assert os.path.exists(vocab_path)
    tokenizer = XLMRobertaTokenizer(vocab_path)
    return tokenizer


class XLMRobertaProcessor(HfTextClassificationProcessor):
    """
    Processor for XLM-RoBERTa model for text classification tasks.
    """

    def __init__(
        self,
        vocab_path: str,
        max_seq_length: Optional[int] = 128,
        source_type_id: Optional[int] = 0,
        target_type_id: Optional[int] = 0,
    ):
        """
        Initializes the XLMRobertaProcessor.

        Args:
            vocab_path (str): Path to the vocabulary file.
            max_seq_length (Optional[int]): Maximum sequence length. Defaults to 128.
            source_type_id (Optional[int]): Source type ID. Defaults to 0.
            target_type_id (Optional[int]): Target type ID. Defaults to 0.
        """
        tokenizer = get_xlm_roberta_tokenizer(
            vocab_path,
        )
        super().__init__(
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
            source_type_id=source_type_id,
            target_type_id=target_type_id,
            position_start_id=self.pad_token_id + 1,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import time
import json
import logging
from unitorch.utils import is_wandb_available, read_json_file
from unitorch.cli.core import CoreConfigureParser

_wandb_available = False

if is_wandb_available():
    import wandb
else:
    wandb = None


def setup(config):
    global _wandb_available
    config.set_default_section("core/wandb")
    team = config.getoption("team", None)
    project = config.getoption("project", None)
    token = config.getoption("token", None)
    name = config.getoption("name", time.strftime("%m/%d/%Y/%H", time.localtime()))
    db_file = config.getoption("db_file", None)

    if wandb is None:
        logging.warning("wandb is not available.")
        return

    rank = int(os.environ.get("RANK", 0))
    if rank > 0:
        logging.warning("wandb is not available in distributed mode.")
        return

    if any(x is None for x in [team, project, token]):
        return

    try:
        wandb.login(key=token, relogin=True)
        if db_file is not None:
            db_info = read_json_file(db_file)
            run_id = db_info.get("run_id", None)
            if run_id is None:
                run_id = wandb.util.generate_id()
                db_info["run_id"] = run_id
                with open(db_file, "w") as f:
                    f.write(json.dumps(db_info))
        else:
            run_id = None
        wandb.init(
            entity=team,
            project=project,
            name=name,
            group=name,
            reinit=True,
            id=run_id,
            resume="allow",
        )
        _wandb_available = True
        logging.info("Login to wandb successfully.")
    except Exception as e:
        logging.error(f"Failed to login to wandb: {e}")
        return


def is_available():
    return _wandb_available


config = getattr(wandb, "config", None)


def log(*args, **kwargs):
    if wandb is None:
        return
    if not is_available():
        return
    wandb.log(*args, **kwargs)


def save(*args, **kwargs):
    if wandb is None:
        return
    if not is_available():
        return
    wandb.save(*args, **kwargs)


def finish(*args, **kwargs):
    if wandb is None:
        return
    if not is_available():
        return
    wandb.finish(*args, **kwargs)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import re
import gc
import json
import logging
import torch
import hashlib
import asyncio
import pandas as pd
from PIL import Image
from torch import autocast
from fastapi import APIRouter, UploadFile, File
from fastapi.responses import StreamingResponse
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from diffusers.utils import numpy_to_pil
from diffusers.models import ControlNetModel
from diffusers.pipelines import (
    StableDiffusionPipeline,
    StableDiffusionImg2ImgPipeline,
    StableDiffusionInpaintPipeline,
    StableDiffusionUpscalePipeline,
    StableDiffusionDepth2ImgPipeline,
    StableVideoDiffusionPipeline,
    StableDiffusionControlNetPipeline,
    StableDiffusionControlNetImg2ImgPipeline,
    StableDiffusionControlNetInpaintPipeline,
)
from unitorch import is_xformers_available
from unitorch.utils import is_remote_url
from unitorch.models.diffusers import GenericStableModel
from unitorch.models.diffusers import StableProcessor

from unitorch.utils import pop_value, nested_dict_value
from unitorch.cli import (
    cached_path,
    register_fastapi,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli import CoreConfigureParser, GenericFastAPI
from unitorch.cli.models.diffusers import (
    pretrained_stable_infos,
    pretrained_stable_extensions_infos,
    load_weight,
)
from unitorch.cli.pipelines import Schedulers


class StableForImage2ImageFastAPIPipeline(GenericStableModel):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        vocab_path: str,
        merge_path: str,
        quant_config_path: Optional[str] = None,
        max_seq_length: Optional[int] = 77,
        pad_token: Optional[str] = "<|endoftext|>",
        weight_path: Optional[Union[str, List[str]]] = None,
        state_dict: Optional[Dict[str, Any]] = None,
        lora_checkpoints: Optional[Union[str, List[str]]] = None,
        lora_weights: Optional[Union[float, List[float]]] = 1.0,
        lora_alphas: Optional[Union[float, List[float]]] = 32,
        device: Optional[Union[str, int]] = "cpu",
        enable_cpu_offload: Optional[bool] = False,
        enable_xformers: Optional[bool] = False,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
        )
        self.processor = StableProcessor(
            vocab_path=vocab_path,
            merge_path=merge_path,
            vae_config_path=vae_config_path,
            max_seq_length=max_seq_length,
            pad_token=pad_token,
        )
        self._device = "cpu" if device == "cpu" else int(device)

        self.from_pretrained(weight_path, state_dict=state_dict)

        self.eval()

        self.pipeline = StableDiffusionImg2ImgPipeline(
            vae=self.vae,
            text_encoder=self.text,
            unet=self.unet,
            scheduler=self.scheduler,
            tokenizer=None,
            safety_checker=None,
            feature_extractor=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)

        if lora_checkpoints is not None:
            self.load_lora_weights(
                lora_checkpoints,
                lora_weights=lora_weights,
                lora_alphas=lora_alphas,
                save_base_state=False,
            )

        self._enable_cpu_offload = enable_cpu_offload
        self._enable_xformers = enable_xformers

        if self._enable_cpu_offload and self._device != "cpu":
            self.pipeline.enable_model_cpu_offload(self._device)
        else:
            self.to(device=self._device)

        if self._enable_xformers and self._device != "cpu":
            assert is_xformers_available(), "Please install xformers first."
            self.pipeline.enable_xformers_memory_efficient_attention()

    @classmethod
    @add_default_section_for_init("core/fastapi/pipeline/stable/image2image")
    def from_core_configure(
        cls,
        config,
        pretrained_name: Optional[str] = None,
        config_path: Optional[str] = None,
        text_config_path: Optional[str] = None,
        vae_config_path: Optional[str] = None,
        scheduler_config_path: Optional[str] = None,
        vocab_path: Optional[str] = None,
        merge_path: Optional[str] = None,
        quant_config_path: Optional[str] = None,
        pretrained_weight_path: Optional[str] = None,
        pad_token: Optional[str] = None,
        device: Optional[str] = None,
        pretrained_lora_names: Optional[Union[str, List[str]]] = None,
        pretrained_lora_weights_path: Optional[Union[str, List[str]]] = None,
        pretrained_lora_weights: Optional[Union[float, List[float]]] = None,
        pretrained_lora_alphas: Optional[Union[float, List[float]]] = None,
        **kwargs,
    ):
        config.set_default_section("core/fastapi/pipeline/stable/image2image")
        pretrained_name = pretrained_name or config.getoption(
            "pretrained_name", "stable-v1.5"
        )
        pretrained_infos = nested_dict_value(pretrained_stable_infos, pretrained_name)

        config_path = config.getoption("config_path", config_path)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_infos, "unet", "config"),
        )
        config_path = cached_path(config_path)

        text_config_path = config.getoption("text_config_path", text_config_path)
        text_config_path = pop_value(
            text_config_path,
            nested_dict_value(pretrained_infos, "text", "config"),
        )
        text_config_path = cached_path(text_config_path)

        vae_config_path = config.getoption("vae_config_path", vae_config_path)
        vae_config_path = pop_value(
            vae_config_path,
            nested_dict_value(pretrained_infos, "vae", "config"),
        )
        vae_config_path = cached_path(vae_config_path)

        scheduler_config_path = config.getoption(
            "scheduler_config_path", scheduler_config_path
        )
        scheduler_config_path = pop_value(
            scheduler_config_path,
            nested_dict_value(pretrained_infos, "scheduler"),
        )
        scheduler_config_path = cached_path(scheduler_config_path)

        vocab_path = config.getoption("vocab_path", vocab_path)
        vocab_path = pop_value(
            vocab_path,
            nested_dict_value(pretrained_infos, "text", "vocab"),
        )
        vocab_path = cached_path(vocab_path)

        merge_path = config.getoption("merge_path", merge_path)
        merge_path = pop_value(
            merge_path,
            nested_dict_value(pretrained_infos, "text", "merge"),
        )
        merge_path = cached_path(merge_path)

        quant_config_path = config.getoption("quant_config_path", quant_config_path)
        if quant_config_path is not None:
            quant_config_path = cached_path(quant_config_path)

        max_seq_length = config.getoption("max_seq_length", 77)
        pad_token = pad_token or config.getoption("pad_token", "<|endoftext|>")
        weight_path = config.getoption("pretrained_weight_path", pretrained_weight_path)
        device = config.getoption("device", "cpu") if device is None else device
        enable_cpu_offload = config.getoption("enable_cpu_offload", True)
        enable_xformers = config.getoption("enable_xformers", True)

        state_dict = None
        if weight_path is None and pretrained_infos is not None:
            state_dict = [
                load_weight(nested_dict_value(pretrained_infos, "unet", "weight")),
                load_weight(nested_dict_value(pretrained_infos, "text", "weight")),
                load_weight(nested_dict_value(pretrained_infos, "vae", "weight")),
            ]

        pretrained_lora_names = pretrained_lora_names or config.getoption(
            "pretrained_lora_names", None
        )
        pretrained_lora_weights = pretrained_lora_weights or config.getoption(
            "pretrained_lora_weights", 1.0
        )
        pretrained_lora_alphas = pretrained_lora_alphas or config.getoption(
            "pretrained_lora_alphas", 32.0
        )

        if (
            isinstance(pretrained_lora_names, str)
            and pretrained_lora_weights_path is None
        ):
            pretrained_lora_weights_path = nested_dict_value(
                pretrained_stable_extensions_infos,
                pretrained_lora_names,
                "lora",
                "weight",
            )
        elif (
            isinstance(pretrained_lora_names, list)
            and pretrained_lora_weights_path is None
        ):
            pretrained_lora_weights_path = [
                nested_dict_value(
                    pretrained_stable_extensions_infos, name, "lora", "weight"
                )
                for name in pretrained_lora_names
            ]
            assert len(pretrained_lora_weights_path) == len(pretrained_lora_weights)
            assert len(pretrained_lora_weights_path) == len(pretrained_lora_alphas)

        lora_weights_path = pretrained_lora_weights_path or config.getoption(
            "pretrained_lora_weights_path", None
        )

        inst = cls(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            vocab_path=vocab_path,
            merge_path=merge_path,
            quant_config_path=quant_config_path,
            pad_token=pad_token,
            max_seq_length=max_seq_length,
            weight_path=weight_path,
            state_dict=state_dict,
            lora_checkpoints=lora_weights_path,
            lora_weights=pretrained_lora_weights,
            lora_alphas=pretrained_lora_alphas,
            device=device,
            enable_cpu_offload=enable_cpu_offload,
            enable_xformers=enable_xformers,
        )
        return inst

    @torch.no_grad()
    @autocast(device_type=("cuda" if torch.cuda.is_available() else "cpu"))
    @add_default_section_for_function("core/fastapi/pipeline/stable/image2image")
    def __call__(
        self,
        text: str,
        image: Image.Image,
        neg_text: Optional[str] = "",
        guidance_scale: Optional[float] = 7.5,
        strength: Optional[float] = 1.0,
        num_timesteps: Optional[int] = 50,
        seed: Optional[int] = 1123,
    ):
        text_inputs = self.processor.text2image_inputs(
            text,
            negative_prompt=neg_text,
        )
        image_inputs = self.processor.image2image_inputs(image)
        inputs = {**text_inputs, **image_inputs}
        self.seed = seed

        inputs = {k: v.unsqueeze(0) if v is not None else v for k, v in inputs.items()}
        inputs = {
            k: v.to(device=self.device) if v is not None else v
            for k, v in inputs.items()
        }

        prompt_outputs = self.get_prompt_outputs(
            input_ids=inputs.get("input_ids"),
            negative_input_ids=inputs.get("negative_input_ids"),
            attention_mask=inputs.get("attention_mask"),
            negative_attention_mask=inputs.get("negative_attention_mask"),
            enable_cpu_offload=self._enable_cpu_offload,
            cpu_offload_device=self._device,
        )

        prompt_embeds = prompt_outputs.prompt_embeds
        negative_prompt_embeds = prompt_outputs.negative_prompt_embeds

        outputs = self.pipeline(
            image=inputs["pixel_values"],
            prompt_embeds=prompt_embeds,
            negative_prompt_embeds=negative_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            num_inference_steps=num_timesteps,
            guidance_scale=guidance_scale,
            strength=strength,
            output_type="np.array",
        )

        images = torch.from_numpy(outputs.images)
        images = numpy_to_pil(images.cpu().numpy())
        return images[0]


@register_fastapi("core/fastapi/stable/image2image")
class StableImage2ImageFastAPI(GenericFastAPI):
    def __init__(self, config: CoreConfigureParser):
        self.config = config
        config.set_default_section(f"core/fastapi/stable/image2image")
        router = config.getoption("router", "/core/fastapi/stable/image2image")
        self._pipe = None
        self._router = APIRouter(prefix=router)
        self._router.add_api_route("/generate", self.serve, methods=["POST"])
        self._router.add_api_route("/status", self.status, methods=["GET"])
        self._router.add_api_route("/start", self.start, methods=["POST"])
        self._router.add_api_route("/stop", self.stop, methods=["GET"])
        self._lock = asyncio.Lock()

    @property
    def router(self):
        return self._router

    def start(
        self,
        pretrained_name: Optional[str] = "stable-v1.5",
        pretrained_lora_names: Optional[Union[str, List[str]]] = None,
        pretrained_lora_weights: Optional[Union[float, List[float]]] = 1.0,
        pretrained_lora_alphas: Optional[Union[float, List[float]]] = 32.0,
    ):
        self._pipe = StableForImage2ImageFastAPIPipeline.from_core_configure(
            self.config,
            pretrained_name=pretrained_name,
            pretrained_lora_names=pretrained_lora_names,
            pretrained_lora_weights=pretrained_lora_weights,
            pretrained_lora_alphas=pretrained_lora_alphas,
        )
        return "start success"

    def stop(self):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None
        return "stop success"

    def status(self):
        return "running" if self._pipe is not None else "stopped"

    async def serve(
        self,
        text: str,
        image: UploadFile,
        guidance_scale: Optional[float] = 7.5,
        strength: Optional[float] = 1.0,
        num_timesteps: Optional[int] = 50,
        seed: Optional[int] = 1123,
    ):
        assert self._pipe is not None
        image_bytes = await image.read()
        image = Image.open(io.BytesIO(image_bytes))
        async with self._lock:
            image = self._pipe(
                text,
                image,
                guidance_scale=guidance_scale,
                strength=strength,
                num_timesteps=num_timesteps,
                seed=seed,
            )
        buffer = io.BytesIO()
        image.save(buffer, format="PNG")

        return StreamingResponse(
            io.BytesIO(buffer.getvalue()),
            media_type="image/png",
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import re
import warnings
import logging
import torch
import torch.nn as nn
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from megatron.core import mpu
from megatron.core import dist_checkpointing
from unitorch.utils import (
    replace,
    load_weight,
    is_diffusers_available,
)
from unitorch.models import CheckpointMixin


class MegatronCheckpointMixin(CheckpointMixin):
    checkpoint_name = "common.pt"

    modules_to_save_checkpoints = []

    def save_checkpoint(
        self,
        ckpt_dir: str,
        weight_name: str = None,
        **kwargs,
    ):
        """
        Save the model's current state as a checkpoint.

        Args:
            ckpt_dir (str): Directory path to save the checkpoint.
            weight_name (str): Name of the weight file.

        Returns:
            None
        """
        if weight_name is None:
            weight_name = self.checkpoint_name

        if not os.path.exists(ckpt_dir):
            os.makedirs(ckpt_dir, exist_ok=True)

        sharded_state_dict = {}
        for module in self.modules_to_save_checkpoints:
            _module = getattr(self, module)
            sharded_state_dict.update(_module.sharded_state_dict())

        dist_checkpointing.save(sharded_state_dict, ckpt_dir)

        logging.info(f"{type(self).__name__} model save checkpoint to {ckpt_dir}")

    def from_checkpoint(
        self,
        ckpt_dir: str,
        weight_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Load model weights from a checkpoint.

        Args:
            ckpt_dir (str): Directory path of the checkpoint.
            weight_name (str): Name of the weight file.

        Returns:
            None
        """
        if weight_name is None:
            weight_name = self.checkpoint_name
        weight_path = os.path.join(ckpt_dir, weight_name)
        if not os.path.exists(weight_path):
            return

        sharded_state_dict = {}
        for module in self.modules_to_save_checkpoints:
            _module = getattr(self, module)
            sharded_state_dict.update(_module.sharded_state_dict())

        checkpoint = dist_checkpointing.load(
            sharded_state_dict=sharded_state_dict, checkpoint_dir=ckpt_dir
        )
        self.load_state_dict(checkpoint, strict=False)
        logging.info(
            f"{type(self).__name__} model load weight from checkpoint {ckpt_dir}"
        )


class GenericMegatronModel(nn.Module, MegatronCheckpointMixin):
    def __init__(self):
        super().__init__()
        pass

    def _init_weights(self, module):
        """
        Initialize the weights of the given module.

        Args:
            module (nn.Module): The module to initialize weights for.
        """
        if isinstance(module, (nn.Linear, nn.Embedding)):
            module.weight.data.normal_(mean=0.0, std=0.02)

        if isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

        if isinstance(module, nn.Linear) and module.bias is not None:
            module.bias.data.zero_()

    def init_weights(self):
        """
        Initialize the weights of the model.
        """
        self.apply(self._init_weights)

    @property
    def dtype(self) -> torch.dtype:
        """
        Returns the data type of the model's parameters.

        Returns:
            torch.dtype: The data type of the model's parameters.
        """
        return next(self.parameters()).dtype

    @property
    def device(self):
        """
        Returns the device of the model's parameters.

        Returns:
            torch.device: The device of the model's parameters.
        """
        return next(self.parameters()).device


from unitorch.models.megatron.modeling_gpt import MegatronGPTForGeneration



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.fastapis.stable.text2image import StableForText2ImageFastAPIPipeline
from unitorch.cli.fastapis.stable.image2image import StableForImage2ImageFastAPIPipeline
from unitorch.cli.fastapis.stable.inpainting import (
    StableForImageInpaintingFastAPIPipeline,
)
from unitorch.cli.fastapis.stable.resolution import (
    StableForImageResolutionFastAPIPipeline,
)
from unitorch.cli.fastapis.stable.image2video import StableForImage2VideoFastAPIPipeline

import unitorch.cli.fastapis.stable.interrogator



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.utils import is_opencv_available

if is_opencv_available():
    from unitorch.cli.pipelines.tools.controlnet.canny import canny
    from unitorch.cli.pipelines.tools.controlnet.depth import depth
    from unitorch.cli.pipelines.tools.controlnet.dwpose import dwpose
    from unitorch.cli.pipelines.tools.controlnet.hed import hed
    from unitorch.cli.pipelines.tools.controlnet.lineart import lineart
    from unitorch.cli.pipelines.tools.controlnet.lineart_anime import lineart_anime

    controlnet_processes = {
        "canny": canny,
        "depth": depth,
        "dwpose": dwpose,
        "hed": hed,
        "lineart": lineart,
        "lineart_anime": lineart_anime,
        "inpainting": lambda x: x,
    }

    adapter_processes = {
        "canny": canny,
        "depth": depth,
        "dwpose": dwpose,
        "hed": hed,
        "lineart": lineart,
        "lineart_anime": lineart_anime,
        "inpainting": lambda x: x,
    }
else:
    controlnet_processes = {}
    adapter_processes = {}



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
import torch
import torch.nn.functional as F
from torch import autocast
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
import diffusers.schedulers as schedulers
from diffusers.pipelines.stable_video_diffusion.pipeline_stable_video_diffusion import (
    randn_tensor,
    _append_dims,
    retrieve_timesteps,
    StableVideoDiffusionPipelineOutput,
)
from transformers import (
    CLIPTextConfig,
    CLIPTextModel,
    CLIPVisionConfig,
    CLIPVisionModelWithProjection,
    CLIPImageProcessor,
)
from diffusers.schedulers import SchedulerMixin
from diffusers.models import (
    ControlNetModel,
    UNet2DModel,
    UNet2DConditionModel,
    AutoencoderKL,
    AutoencoderKLTemporalDecoder,
    UNetSpatioTemporalConditionModel,
    MultiAdapter,
    T2IAdapter,
)
from diffusers.pipelines.controlnet import MultiControlNetModel
from diffusers.pipelines import (
    StableDiffusionPipeline,
    StableDiffusionImg2ImgPipeline,
    StableDiffusionInpaintPipeline,
    StableDiffusionUpscalePipeline,
    StableDiffusionDepth2ImgPipeline,
    StableVideoDiffusionPipeline,
)
from unitorch.models import (
    GenericModel,
    GenericOutputs,
    QuantizationConfig,
    QuantizationMixin,
)
from unitorch.models.peft import PeftWeightLoaderMixin


def compute_snr(timesteps, noise_scheduler):
    """
    Computes SNR as per https://github.com/TiankaiHang/Min-SNR-Diffusion-Training/blob/521b624bd70c67cee4bdf49225915f5945a872e3/guided_diffusion/gaussian_diffusion.py#L847-L849
    """
    alphas_cumprod = noise_scheduler.alphas_cumprod
    sqrt_alphas_cumprod = alphas_cumprod**0.5
    sqrt_one_minus_alphas_cumprod = (1.0 - alphas_cumprod) ** 0.5
    # Expand the tensors.
    # Adapted from https://github.com/TiankaiHang/Min-SNR-Diffusion-Training/blob/521b624bd70c67cee4bdf49225915f5945a872e3/guided_diffusion/gaussian_diffusion.py#L1026
    sqrt_alphas_cumprod = sqrt_alphas_cumprod.to(device=timesteps.device)[
        timesteps
    ].float()
    while len(sqrt_alphas_cumprod.shape) < len(timesteps.shape):
        sqrt_alphas_cumprod = sqrt_alphas_cumprod[..., None]
    alpha = sqrt_alphas_cumprod.expand(timesteps.shape)

    sqrt_one_minus_alphas_cumprod = sqrt_one_minus_alphas_cumprod.to(
        device=timesteps.device
    )[timesteps].float()
    while len(sqrt_one_minus_alphas_cumprod.shape) < len(timesteps.shape):
        sqrt_one_minus_alphas_cumprod = sqrt_one_minus_alphas_cumprod[..., None]
    sigma = sqrt_one_minus_alphas_cumprod.expand(timesteps.shape)

    # Compute SNR
    snr = (alpha / sigma) ** 2
    return snr


class GenericStableModel(GenericModel, QuantizationMixin, PeftWeightLoaderMixin):
    prefix_keys_in_state_dict = {
        # unet weights
        "^conv_in.*": "unet.",
        "^conv_norm_out.*": "unet.",
        "^conv_out.*": "unet.",
        "^time_embedding.*": "unet.",
        "^class_embedding.*": "unet.",
        "^up_blocks.*": "unet.",
        "^mid_block.*": "unet.",
        "^down_blocks.*": "unet.",
        # text weights
        "^text_model.*": "text.",
        # vae weights
        "^encoder.*": "vae.",
        "^decoder.*": "vae.",
        "^post_quant_conv.*": "vae.",
        "^quant_conv.*": "vae.",
    }

    replace_keys_in_state_dict = {
        "\.query\.": ".to_q.",
        "\.key\.": ".to_k.",
        "\.value\.": ".to_v.",
        "\.proj_attn\.": ".to_out.0.",
    }

    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        controlnet_configs_path: Union[str, List[str]] = None,
        inpainting_controlnet_config_path: Union[str] = None,
        adapter_configs_path: Union[str, List[str]] = None,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        freeze_unet_encoder: Optional[bool] = False,
        snr_gamma: Optional[float] = 5.0,
        seed: Optional[int] = 1123,
    ):
        super().__init__()
        self.seed = seed
        self.num_train_timesteps = num_train_timesteps
        self.num_infer_timesteps = num_infer_timesteps
        self.image_size = image_size
        self.snr_gamma = snr_gamma

        config_dict = json.load(open(config_path))
        if image_size is not None:
            config_dict.update({"sample_size": image_size})
        if in_channels is not None:
            config_dict.update({"in_channels": in_channels})
        if out_channels is not None:
            config_dict.update({"out_channels": out_channels})
        self.unet = UNet2DConditionModel.from_config(config_dict)

        text_config = CLIPTextConfig.from_json_file(text_config_path)
        self.text = CLIPTextModel(text_config)

        vae_config_dict = json.load(open(vae_config_path))
        self.vae = AutoencoderKL.from_config(vae_config_dict)

        if isinstance(controlnet_configs_path, str):
            controlnet_configs_path = [controlnet_configs_path]
        if isinstance(inpainting_controlnet_config_path, str):
            controlnet_configs_path += [inpainting_controlnet_config_path]

        if isinstance(controlnet_configs_path, list):
            if len(controlnet_configs_path) == 0:
                controlnet_configs_path = None
            elif len(controlnet_configs_path) == 1:
                controlnet_configs_path = controlnet_configs_path[0]

        if isinstance(controlnet_configs_path, list):
            controlnets = []
            for controlnet_config_path in controlnet_configs_path:
                controlnet_config_dict = json.load(open(controlnet_config_path))
                controlnets.append(ControlNetModel.from_config(controlnet_config_dict))
            self.num_controlnets = len(controlnets)
            self.controlnet = MultiControlNetModel(
                controlnets=controlnets,
            )
        elif isinstance(controlnet_configs_path, str):
            controlnet_config_dict = json.load(open(controlnet_configs_path))
            self.controlnet = ControlNetModel.from_config(controlnet_config_dict)
            self.num_controlnets = 1
        else:
            self.controlnet = None
            self.num_controlnets = 0

        if isinstance(adapter_configs_path, list):
            if len(adapter_configs_path) == 0:
                adapter_configs_path = None
            elif len(adapter_configs_path) == 1:
                adapter_configs_path = adapter_configs_path[0]

        if isinstance(adapter_configs_path, list):
            adapters = []
            for adapter_config_path in adapter_configs_path:
                adapter_config_dict = json.load(open(adapter_config_path))
                adapters.append(T2IAdapter.from_config(adapter_config_dict))
            self.num_adapters = len(adapters)
            self.adapter = MultiAdapter(
                adapters=adapters,
            )
        elif isinstance(adapter_configs_path, str):
            adapter_config_dict = json.load(open(adapter_configs_path))
            self.adapter = T2IAdapter.from_config(adapter_config_dict)
            self.num_adapters = 1
        else:
            self.adapter = None
            self.num_adapters = 0

        scheduler_config_dict = json.load(open(scheduler_config_path))
        scheduler_class_name = scheduler_config_dict.get("_class_name", "DDPMScheduler")
        assert hasattr(schedulers, scheduler_class_name)
        scheduler_class = getattr(schedulers, scheduler_class_name)
        assert issubclass(scheduler_class, SchedulerMixin)
        scheduler_config_dict["num_train_timesteps"] = num_train_timesteps
        self.scheduler = scheduler_class.from_config(scheduler_config_dict)

        self.vae_scale_factor = 2 ** (len(self.vae.config.block_out_channels) - 1)

        if freeze_vae_encoder:
            for param in self.vae.parameters():
                param.requires_grad = False

        if freeze_text_encoder:
            for param in self.text.parameters():
                param.requires_grad = False

        if freeze_unet_encoder:
            for param in self.unet.parameters():
                param.requires_grad = False

        if quant_config_path is not None:
            self.quant_config = QuantizationConfig.from_json_file(quant_config_path)
            self.quantize(self.quant_config, ignore_modules=["lm_head", "unet", "vae"])

    def get_prompt_outputs(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        enable_cpu_offload: Optional[bool] = False,
        cpu_offload_device: Optional[str] = "cpu",
    ):
        if enable_cpu_offload:
            self.text.to(cpu_offload_device)
            input_ids = input_ids.to(cpu_offload_device)
            negative_input_ids = negative_input_ids.to(cpu_offload_device)
        prompt_embeds = self.text(
            input_ids,
            # attention_mask,
        )[0]
        negative_prompt_embeds = self.text(
            negative_input_ids,
            # negative_attention_mask,
        )[0]
        if enable_cpu_offload:
            self.text.to("cpu")
        return GenericOutputs(
            prompt_embeds=prompt_embeds.to("cpu")
            if enable_cpu_offload
            else prompt_embeds,
            negative_prompt_embeds=negative_prompt_embeds.to("cpu")
            if enable_cpu_offload
            else negative_prompt_embeds,
        )


class StableForText2ImageGeneration(GenericStableModel):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        snr_gamma: Optional[float] = 5.0,
        seed: Optional[int] = 1123,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            snr_gamma=snr_gamma,
            seed=seed,
        )

        self.pipeline = StableDiffusionPipeline(
            vae=self.vae,
            text_encoder=self.text,
            unet=self.unet,
            scheduler=self.scheduler,
            tokenizer=None,
            safety_checker=None,
            feature_extractor=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)

    def forward(
        self,
        pixel_values: torch.Tensor,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
    ):
        latents = self.vae.encode(pixel_values).latent_dist.sample()
        latents = latents * self.vae.config.scaling_factor
        noise = torch.randn(latents.shape).to(latents.device)
        batch = latents.size(0)

        timesteps = torch.randint(
            0,
            self.scheduler.config.num_train_timesteps,
            (batch,),
            device=pixel_values.device,
        ).long()

        noise_latents = self.scheduler.add_noise(
            latents,
            noise,
            timesteps,
        )

        encoder_hidden_states = self.text(input_ids)[0]
        # encoder_hidden_states = self.text(input_ids, attention_mask)[0]
        outputs = self.unet(
            noise_latents,
            timesteps,
            encoder_hidden_states,
        ).sample
        if self.scheduler.config.prediction_type == "v_prediction":
            noise = self.scheduler.get_velocity(latents, noise, timesteps)
        if self.snr_gamma > 0:
            snr = compute_snr(timesteps, self.scheduler)
            base_weight = (
                torch.stack(
                    [snr, self.snr_gamma * torch.ones_like(timesteps)], dim=1
                ).min(dim=1)[0]
                / snr
            )

            if self.scheduler.config.prediction_type == "v_prediction":
                mse_loss_weights = base_weight + 1
            else:
                mse_loss_weights = base_weight
            loss = F.mse_loss(outputs, noise, reduction="none")
            loss = loss.mean(dim=list(range(1, len(loss.shape)))) * mse_loss_weights
            loss = loss.mean()
        else:
            loss = F.mse_loss(outputs, noise, reduction="mean")
        return loss

    def generate(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        height: Optional[int] = 512,
        width: Optional[int] = 512,
        guidance_scale: Optional[float] = 7.5,
    ):
        outputs = self.get_prompt_outputs(
            input_ids=input_ids,
            negative_input_ids=negative_input_ids,
            attention_mask=attention_mask,
            negative_attention_mask=negative_attention_mask,
        )

        images = self.pipeline(
            prompt_embeds=outputs.prompt_embeds,
            negative_prompt_embeds=outputs.negative_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            num_inference_steps=self.num_infer_timesteps,
            height=height,
            width=width,
            guidance_scale=guidance_scale,
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))


class StableForImage2ImageGeneration(GenericStableModel):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        snr_gamma: Optional[float] = 5.0,
        seed: Optional[int] = 1123,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            snr_gamma=snr_gamma,
            seed=seed,
        )

        self.pipeline = StableDiffusionImg2ImgPipeline(
            vae=self.vae,
            text_encoder=self.text,
            unet=self.unet,
            scheduler=self.scheduler,
            tokenizer=None,
            safety_checker=None,
            feature_extractor=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)

    def forward(
        self,
    ):
        raise NotImplementedError

    def generate(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        strength: Optional[float] = 1.0,
        guidance_scale: Optional[float] = 7.5,
    ):
        outputs = self.get_prompt_outputs(
            input_ids=input_ids,
            negative_input_ids=negative_input_ids,
            attention_mask=attention_mask,
            negative_attention_mask=negative_attention_mask,
        )

        images = self.pipeline(
            image=pixel_values,
            prompt_embeds=outputs.prompt_embeds,
            negative_prompt_embeds=outputs.negative_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            num_inference_steps=self.num_infer_timesteps,
            strength=strength,
            guidance_scale=guidance_scale,
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))


class StableForImageInpainting(GenericStableModel):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        snr_gamma: Optional[float] = 5.0,
        seed: Optional[int] = 1123,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            snr_gamma=snr_gamma,
            seed=seed,
        )

        self.pipeline = StableDiffusionInpaintPipeline(
            vae=self.vae,
            text_encoder=self.text,
            unet=self.unet,
            scheduler=self.scheduler,
            tokenizer=None,
            safety_checker=None,
            feature_extractor=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)
        self.num_channels_unet = self.unet.config.in_channels

    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        pixel_masks: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
    ):
        latents = self.vae.encode(pixel_values).latent_dist.sample()
        latents = latents * self.vae.config.scaling_factor

        noise = torch.randn(latents.shape).to(latents.device)
        batch = latents.size(0)

        timesteps = torch.randint(
            0,
            self.scheduler.config.num_train_timesteps,
            (batch,),
            device=pixel_values.device,
        ).long()

        noise_latents = self.scheduler.add_noise(
            latents,
            noise,
            timesteps,
        )

        if self.num_channels_unet == 9:
            masked_pixel_values = pixel_values.clone()
            masked_pixel_masks = pixel_masks.clone()
            masked_pixel_masks = masked_pixel_masks.expand_as(masked_pixel_values)
            masked_pixel_values[masked_pixel_masks > 0.5] = 0.0
            masked_latents = self.vae.encode(masked_pixel_values).latent_dist.sample()
            masked_latents = masked_latents * self.vae.config.scaling_factor

            pixel_masks = torch.nn.functional.interpolate(
                pixel_masks, size=latents.shape[-2:], mode="nearest"
            )
            latent_model_input = torch.cat(
                [noise_latents, pixel_masks, masked_latents], dim=1
            )
        else:
            latent_model_input = noise_latents

        encoder_hidden_states = self.text(input_ids)[0]
        # encoder_hidden_states = self.text(input_ids, attention_mask)[0]
        outputs = self.unet(
            latent_model_input,
            timesteps,
            encoder_hidden_states,
        ).sample
        if self.scheduler.config.prediction_type == "v_prediction":
            noise = self.scheduler.get_velocity(latents, noise, timesteps)
        if self.snr_gamma > 0:
            snr = compute_snr(timesteps, self.scheduler)
            base_weight = (
                torch.stack(
                    [snr, self.snr_gamma * torch.ones_like(timesteps)], dim=1
                ).min(dim=1)[0]
                / snr
            )

            if self.scheduler.config.prediction_type == "v_prediction":
                mse_loss_weights = base_weight + 1
            else:
                mse_loss_weights = base_weight
            loss = F.mse_loss(outputs, noise, reduction="none")
            loss = loss.mean(dim=list(range(1, len(loss.shape)))) * mse_loss_weights
            loss = loss.mean()
        else:
            loss = F.mse_loss(outputs, noise, reduction="mean")
        return loss

    def generate(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        pixel_masks: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        strength: Optional[float] = 1.0,
        guidance_scale: Optional[float] = 7.5,
    ):
        outputs = self.get_prompt_outputs(
            input_ids=input_ids,
            negative_input_ids=negative_input_ids,
            attention_mask=attention_mask,
            negative_attention_mask=negative_attention_mask,
        )

        images = self.pipeline(
            image=pixel_values,
            mask_image=pixel_masks,
            prompt_embeds=outputs.prompt_embeds,
            negative_prompt_embeds=outputs.negative_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            num_inference_steps=self.num_infer_timesteps,
            width=pixel_values.size(-1),
            height=pixel_values.size(-2),
            strength=strength,
            guidance_scale=guidance_scale,
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))


class StableForImageResolution(GenericStableModel):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        snr_gamma: Optional[float] = 5.0,
        noise_level: Optional[int] = 20,
        seed: Optional[int] = 1123,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            snr_gamma=snr_gamma,
            seed=seed,
        )

        self.pipeline = StableDiffusionUpscalePipeline(
            vae=self.vae,
            text_encoder=self.text,
            unet=self.unet,
            scheduler=self.scheduler,
            low_res_scheduler=self.scheduler,
            tokenizer=None,
            safety_checker=None,
            feature_extractor=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)
        self.num_channels_unet = self.unet.config.in_channels
        self.noise_level = noise_level
        self.low_res_scheduler = schedulers.DDPMScheduler.from_config(
            {
                "beta_end": 0.02,
                "beta_schedule": "scaled_linear",
                "beta_start": 0.0001,
                "clip_sample": True,
                "num_train_timesteps": 1000,
                "prediction_type": "epsilon",
                "trained_betas": None,
                "variance_type": "fixed_small",
            }
        )

    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        low_res_pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
    ):
        with autocast(
            device_type=("cuda" if torch.cuda.is_available() else "cpu"),
            dtype=torch.float32,
        ):
            latents = self.vae.encode(pixel_values).latent_dist.sample()
            latents = latents * self.vae.config.scaling_factor

        noise = torch.randn(low_res_pixel_values.shape).to(low_res_pixel_values.device)
        noise_level = torch.tensor(
            [self.noise_level], dtype=torch.long, device=pixel_values.device
        )
        low_res_pixel_values = self.low_res_scheduler.add_noise(
            low_res_pixel_values, noise, noise_level
        )

        noise = torch.randn(latents.shape).to(latents.device)
        batch = latents.size(0)

        timesteps = torch.randint(
            0,
            self.scheduler.config.num_train_timesteps,
            (batch,),
            device=pixel_values.device,
        ).long()

        noise_latents = self.scheduler.add_noise(
            latents,
            noise,
            timesteps,
        )

        if self.num_channels_unet == 7:
            latent_model_input = torch.cat([noise_latents, low_res_pixel_values], dim=1)
        else:
            latent_model_input = noise_latents

        encoder_hidden_states = self.text(input_ids)[0]
        # encoder_hidden_states = self.text(input_ids, attention_mask)[0]
        outputs = self.unet(
            latent_model_input,
            timesteps,
            encoder_hidden_states,
            class_labels=noise_level,
        ).sample
        if self.scheduler.config.prediction_type == "v_prediction":
            noise = self.scheduler.get_velocity(latents, noise, timesteps)
        if self.snr_gamma > 0:
            snr = compute_snr(timesteps, self.scheduler)
            base_weight = (
                torch.stack(
                    [snr, self.snr_gamma * torch.ones_like(timesteps)], dim=1
                ).min(dim=1)[0]
                / snr
            )

            if self.scheduler.config.prediction_type == "v_prediction":
                mse_loss_weights = base_weight + 1
            else:
                mse_loss_weights = base_weight
            loss = F.mse_loss(outputs, noise, reduction="none")
            loss = loss.mean(dim=list(range(1, len(loss.shape)))) * mse_loss_weights
            loss = loss.mean()
        else:
            loss = F.mse_loss(outputs, noise, reduction="mean")
        return loss

    def generate(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        guidance_scale: Optional[float] = 9.0,
        noise_level: Optional[int] = 20,
    ):
        outputs = self.get_prompt_outputs(
            input_ids=input_ids,
            negative_input_ids=negative_input_ids,
            attention_mask=attention_mask,
            negative_attention_mask=negative_attention_mask,
        )

        images = self.pipeline(
            image=pixel_values,
            prompt_embeds=outputs.prompt_embeds,
            negative_prompt_embeds=outputs.negative_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            num_inference_steps=self.num_infer_timesteps,
            guidance_scale=guidance_scale,
            noise_level=noise_level,
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))


class StableVideoDiffusionPipelineV2(StableVideoDiffusionPipeline):
    def __call__(
        self,
        image: Union[Image.Image, List[Image.Image], torch.Tensor],
        vae_image: Union[Image.Image, List[Image.Image], torch.Tensor],
        height: int = 576,
        width: int = 1024,
        num_frames: Optional[int] = None,
        num_inference_steps: int = 25,
        sigmas: Optional[List[float]] = None,
        min_guidance_scale: float = 1.0,
        max_guidance_scale: float = 3.0,
        fps: int = 7,
        motion_bucket_id: int = 127,
        noise_aug_strength: float = 0.02,
        decode_chunk_size: Optional[int] = None,
        num_videos_per_prompt: Optional[int] = 1,
        generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None,
        latents: Optional[torch.Tensor] = None,
        output_type: Optional[str] = "pil",
        callback_on_step_end: Optional[Callable[[int, int, Dict], None]] = None,
        callback_on_step_end_tensor_inputs: List[str] = ["latents"],
        return_dict: bool = True,
    ):
        height = height or self.unet.config.sample_size * self.vae_scale_factor
        width = width or self.unet.config.sample_size * self.vae_scale_factor

        num_frames = (
            num_frames if num_frames is not None else self.unet.config.num_frames
        )
        decode_chunk_size = (
            decode_chunk_size if decode_chunk_size is not None else num_frames
        )

        # 1. Check inputs. Raise error if not correct
        self.check_inputs(image, height, width)

        # 2. Define call parameters
        if isinstance(image, Image.Image):
            batch_size = 1
        elif isinstance(image, list):
            batch_size = len(image)
        else:
            batch_size = image.shape[0]
        device = self._execution_device
        # here `guidance_scale` is defined analog to the guidance weight `w` of equation (2)
        # of the Imagen paper: https://arxiv.org/pdf/2205.11487.pdf . `guidance_scale = 1`
        # corresponds to doing no classifier free guidance.
        self._guidance_scale = max_guidance_scale

        # 3. Encode input image
        # image_embeddings = self._encode_image(
        #     image, device, num_videos_per_prompt, self.do_classifier_free_guidance
        # )
        image_embeddings = self.image_encoder(image).image_embeds
        image_embeddings = image_embeddings.unsqueeze(1)

        # duplicate image embeddings for each generation per prompt, using mps friendly method
        bs_embed, seq_len, _ = image_embeddings.shape
        image_embeddings = image_embeddings.repeat(1, num_videos_per_prompt, 1)
        image_embeddings = image_embeddings.view(
            bs_embed * num_videos_per_prompt, seq_len, -1
        )

        if self.do_classifier_free_guidance:
            negative_image_embeddings = torch.zeros_like(image_embeddings)

            # For classifier free guidance, we need to do two forward passes.
            # Here we concatenate the unconditional and text embeddings into a single batch
            # to avoid doing two forward passes
            image_embeddings = torch.cat([negative_image_embeddings, image_embeddings])

        # NOTE: Stable Video Diffusion was conditioned on fps - 1, which is why it is reduced here.
        # See: https://github.com/Stability-AI/generative-models/blob/ed0997173f98eaf8f4edf7ba5fe8f15c6b877fd3/scripts/sampling/simple_video_sample.py#L188
        fps = fps - 1

        # 4. Encode input image using VAE
        # image = self.video_processor.preprocess(image, height=height, width=width).to(device)
        noise = randn_tensor(
            vae_image.shape, generator=generator, device=device, dtype=vae_image.dtype
        )
        vae_image = vae_image + noise_aug_strength * noise

        needs_upcasting = (
            self.vae.dtype == torch.float16 and self.vae.config.force_upcast
        )
        if needs_upcasting:
            self.vae.to(dtype=torch.float32)

        image_latents = self._encode_vae_image(
            vae_image,
            device=device,
            num_videos_per_prompt=num_videos_per_prompt,
            do_classifier_free_guidance=self.do_classifier_free_guidance,
        )
        image_latents = image_latents.to(image_embeddings.dtype)

        # cast back to fp16 if needed
        if needs_upcasting:
            self.vae.to(dtype=torch.float16)

        # Repeat the image latents for each frame so we can concatenate them with the noise
        # image_latents [batch, channels, height, width] ->[batch, num_frames, channels, height, width]
        image_latents = image_latents.unsqueeze(1).repeat(1, num_frames, 1, 1, 1)

        # 5. Get Added Time IDs
        added_time_ids = self._get_add_time_ids(
            fps,
            motion_bucket_id,
            noise_aug_strength,
            image_embeddings.dtype,
            batch_size,
            num_videos_per_prompt,
            self.do_classifier_free_guidance,
        )
        added_time_ids = added_time_ids.to(device)

        # 6. Prepare timesteps
        timesteps, num_inference_steps = retrieve_timesteps(
            self.scheduler, num_inference_steps, device, None, sigmas
        )

        # 7. Prepare latent variables
        num_channels_latents = self.unet.config.in_channels
        latents = self.prepare_latents(
            batch_size * num_videos_per_prompt,
            num_frames,
            num_channels_latents,
            height,
            width,
            image_embeddings.dtype,
            device,
            generator,
            latents,
        )

        # 8. Prepare guidance scale
        guidance_scale = torch.linspace(
            min_guidance_scale, max_guidance_scale, num_frames
        ).unsqueeze(0)
        guidance_scale = guidance_scale.to(device, latents.dtype)
        guidance_scale = guidance_scale.repeat(batch_size * num_videos_per_prompt, 1)
        guidance_scale = _append_dims(guidance_scale, latents.ndim)

        self._guidance_scale = guidance_scale

        # 9. Denoising loop
        num_warmup_steps = len(timesteps) - num_inference_steps * self.scheduler.order
        self._num_timesteps = len(timesteps)
        with self.progress_bar(total=num_inference_steps) as progress_bar:
            for i, t in enumerate(timesteps):
                # expand the latents if we are doing classifier free guidance
                latent_model_input = (
                    torch.cat([latents] * 2)
                    if self.do_classifier_free_guidance
                    else latents
                )
                latent_model_input = self.scheduler.scale_model_input(
                    latent_model_input, t
                )

                # Concatenate image_latents over channels dimension
                latent_model_input = torch.cat(
                    [latent_model_input, image_latents], dim=2
                )

                # predict the noise residual
                noise_pred = self.unet(
                    latent_model_input,
                    t,
                    encoder_hidden_states=image_embeddings,
                    added_time_ids=added_time_ids,
                    return_dict=False,
                )[0]

                # perform guidance
                if self.do_classifier_free_guidance:
                    noise_pred_uncond, noise_pred_cond = noise_pred.chunk(2)
                    noise_pred = noise_pred_uncond + self.guidance_scale * (
                        noise_pred_cond - noise_pred_uncond
                    )

                # compute the previous noisy sample x_t -> x_t-1
                latents = self.scheduler.step(noise_pred, t, latents).prev_sample

                if callback_on_step_end is not None:
                    callback_kwargs = {}
                    for k in callback_on_step_end_tensor_inputs:
                        callback_kwargs[k] = locals()[k]
                    callback_outputs = callback_on_step_end(self, i, t, callback_kwargs)

                    latents = callback_outputs.pop("latents", latents)

                if i == len(timesteps) - 1 or (
                    (i + 1) > num_warmup_steps and (i + 1) % self.scheduler.order == 0
                ):
                    progress_bar.update()

        if not output_type == "latent":
            # cast back to fp16 if needed
            if needs_upcasting:
                self.vae.to(dtype=torch.float16)
            frames = self.decode_latents(latents, num_frames, decode_chunk_size)
            frames = self.video_processor.postprocess_video(
                video=frames, output_type=output_type
            )
        else:
            frames = latents

        self.maybe_free_model_hooks()

        if not return_dict:
            return frames

        return StableVideoDiffusionPipelineOutput(frames=frames)


class StableForImage2VideoGeneration(GenericModel):
    prefix_keys_in_state_dict = {
        # unet weights
        "^add_embedding.*": "unet.",
        "^conv_in.*": "unet.",
        "^conv_norm_out.*": "unet.",
        "^conv_out.*": "unet.",
        "^time_embedding.*": "unet.",
        "^class_embedding.*": "unet.",
        "^up_blocks.*": "unet.",
        "^mid_block.*": "unet.",
        "^down_blocks.*": "unet.",
        # vision weights
        "^vision_model.*": "image.",
        "^visual_projection.*": "image.",
        # vae weights
        "^encoder.*": "vae.",
        "^decoder.*": "vae.",
        "^post_quant_conv.*": "vae.",
        "^quant_conv.*": "vae.",
    }

    replace_keys_in_state_dict = {
        "\.query\.": ".to_q.",
        "\.key\.": ".to_k.",
        "\.value\.": ".to_v.",
        "\.proj_attn\.": ".to_out.0.",
    }

    def __init__(
        self,
        config_path: str,
        image_config_path: str,
        image_process_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 25,
        freeze_vae_encoder: Optional[bool] = True,
        snr_gamma: Optional[float] = 5.0,
        seed: Optional[int] = 1123,
    ):
        super().__init__()
        self.seed = seed
        self.num_train_timesteps = num_train_timesteps
        self.num_infer_timesteps = num_infer_timesteps

        config_dict = json.load(open(config_path))
        if image_size is not None:
            config_dict.update({"sample_size": image_size})
        if in_channels is not None:
            config_dict.update({"in_channels": in_channels})
        if out_channels is not None:
            config_dict.update({"out_channels": out_channels})
        self.unet = UNetSpatioTemporalConditionModel.from_config(config_dict)

        image_config = CLIPVisionConfig.from_json_file(image_config_path)
        self.image = CLIPVisionModelWithProjection(image_config)
        self.image_processor = CLIPImageProcessor.from_json_file(
            image_process_config_path
        )

        vae_config_dict = json.load(open(vae_config_path))
        self.vae = AutoencoderKLTemporalDecoder.from_config(vae_config_dict)

        scheduler_config_dict = json.load(open(scheduler_config_path))
        scheduler_class_name = scheduler_config_dict.get("_class_name", "DDPMScheduler")
        assert hasattr(schedulers, scheduler_class_name)
        scheduler_class = getattr(schedulers, scheduler_class_name)
        assert issubclass(scheduler_class, SchedulerMixin)
        scheduler_config_dict["num_train_timesteps"] = num_train_timesteps
        self.scheduler = scheduler_class.from_config(scheduler_config_dict)

        self.vae_scale_factor = 2 ** (len(self.vae.config.block_out_channels) - 1)

        self.pipeline = StableVideoDiffusionPipelineV2(
            vae=self.vae,
            image_encoder=self.image,
            unet=self.unet,
            scheduler=self.scheduler,
            feature_extractor=self.image_processor,
        )
        self.pipeline.set_progress_bar_config(disable=True)

        if freeze_vae_encoder:
            for param in self.vae.parameters():
                param.requires_grad = False

        if quant_config_path is not None:
            self.quant_config = QuantizationConfig.from_json_file(quant_config_path)
            self.quantize(self.quant_config, ignore_modules=["lm_head", "unet", "vae"])

    def forward(
        self,
    ):
        raise NotImplementedError

    def generate(
        self,
        pixel_values: torch.Tensor,
        vae_pixel_values: torch.Tensor,
        height: Optional[int] = 576,
        width: Optional[int] = 1024,
        num_frames: Optional[int] = 30,
        fps: Optional[int] = 6,
        min_guidance_scale: Optional[float] = 1.0,
        max_guidance_scale: Optional[float] = 2.5,
        motion_bucket_id: Optional[int] = 127,
        decode_chunk_size: Optional[int] = 8,
    ):
        frames = self.pipeline(
            image=pixel_values,
            vae_image=vae_pixel_values,
            height=height,
            width=width,
            num_frames=num_frames,
            min_guidance_scale=min_guidance_scale,
            max_guidance_scale=max_guidance_scale,
            fps=fps,
            motion_bucket_id=motion_bucket_id,
            decode_chunk_size=decode_chunk_size,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            num_inference_steps=self.num_infer_timesteps,
            output_type="pt",
        ).frames

        return GenericOutputs(frames=frames)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import math
import torch.optim as optim
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union


class CosineWarmupScheduler(optim.lr_scheduler.LambdaLR):
    def __init__(
        self,
        optimizer: optim.Optimizer,
        num_warmup_steps: int,
        num_training_steps: int,
        num_cycles: Optional[float] = 0.5,
        last_epoch: Optional[int] = -1,
    ):
        def lr_lambda(current_step):
            if current_step < num_warmup_steps:
                return float(current_step) / float(max(1, num_warmup_steps))
            progress = float(current_step - num_warmup_steps) / float(
                max(1, num_training_steps - num_warmup_steps)
            )
            return max(
                0.0,
                0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress)),
            )

        super().__init__(optimizer, lr_lambda, last_epoch)


class LinearWarmupScheduler(optim.lr_scheduler.LambdaLR):
    def __init__(
        self,
        optimizer: optim.Optimizer,
        num_warmup_steps: int,
        num_training_steps: int,
        last_epoch: Optional[int] = -1,
    ):
        def lr_lambda(current_step: int):
            if current_step < num_warmup_steps:
                return float(current_step) / float(max(1, num_warmup_steps))
            return max(
                0.0,
                float(num_training_steps - current_step)
                / float(max(1, num_training_steps - num_warmup_steps)),
            )

        super().__init__(optimizer, lr_lambda, last_epoch)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import cv2
import torch
import numpy as np
import torch.nn as nn
from PIL import Image
from torchvision.transforms import ToTensor, ToPILImage
from unitorch.models import GenericModel
from unitorch.cli import hf_endpoint_url


class DoubleConvBlock(torch.nn.Module):
    def __init__(self, input_channel, output_channel, layer_number):
        super().__init__()
        self.convs = torch.nn.Sequential()
        self.convs.append(
            torch.nn.Conv2d(
                in_channels=input_channel,
                out_channels=output_channel,
                kernel_size=(3, 3),
                stride=(1, 1),
                padding=1,
            )
        )
        for i in range(1, layer_number):
            self.convs.append(
                torch.nn.Conv2d(
                    in_channels=output_channel,
                    out_channels=output_channel,
                    kernel_size=(3, 3),
                    stride=(1, 1),
                    padding=1,
                )
            )
        self.projection = torch.nn.Conv2d(
            in_channels=output_channel,
            out_channels=1,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=0,
        )

    def __call__(self, x, down_sampling=False):
        h = x
        if down_sampling:
            h = torch.nn.functional.max_pool2d(h, kernel_size=(2, 2), stride=(2, 2))
        for conv in self.convs:
            h = conv(h)
            h = torch.nn.functional.relu(h)
        return h, self.projection(h)


class ControlNetHED_Apache2(GenericModel):
    def __init__(self):
        super().__init__()
        self.norm = torch.nn.Parameter(torch.zeros(size=(1, 3, 1, 1)))
        self.block1 = DoubleConvBlock(
            input_channel=3, output_channel=64, layer_number=2
        )
        self.block2 = DoubleConvBlock(
            input_channel=64, output_channel=128, layer_number=2
        )
        self.block3 = DoubleConvBlock(
            input_channel=128, output_channel=256, layer_number=3
        )
        self.block4 = DoubleConvBlock(
            input_channel=256, output_channel=512, layer_number=3
        )
        self.block5 = DoubleConvBlock(
            input_channel=512, output_channel=512, layer_number=3
        )

    def forward(self, x):
        h = x - self.norm
        h, projection1 = self.block1(h)
        h, projection2 = self.block2(h, down_sampling=True)
        h, projection3 = self.block3(h, down_sampling=True)
        h, projection4 = self.block4(h, down_sampling=True)
        h, projection5 = self.block5(h, down_sampling=True)
        return projection1, projection2, projection3, projection4, projection5


class HedPipeline(ControlNetHED_Apache2):
    def __init__(self):
        super().__init__()
        self.from_pretrained(
            hf_endpoint_url("/lllyasviel/Annotators/resolve/main/ControlNetHED.pth")
        )
        self.preprocess = ToTensor()
        self.postprocess = ToPILImage()
        self.eval()

    @torch.no_grad()
    def __call__(self, image: Image.Image):
        image = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)
        H, W, C = image.shape
        pixel_values = torch.from_numpy(image).float().permute(2, 0, 1).unsqueeze(0)
        pixel_values = pixel_values.to(self.device)
        edges = self.forward(pixel_values)
        edges = [e.detach().cpu().numpy().astype(np.float32)[0, 0] for e in edges]
        edges = [cv2.resize(e, (W, H), interpolation=cv2.INTER_LINEAR) for e in edges]
        edges = np.stack(edges, axis=2)
        edge = 1 / (1 + np.exp(-np.mean(edges, axis=2).astype(np.float64)))
        edge = (edge * 255.0).clip(0, 255).astype(np.uint8)
        return Image.fromarray(cv2.cvtColor(edge, cv2.COLOR_BGR2RGB))


hed_pipe = None


def hed(image: Image.Image):
    global hed_pipe
    if hed_pipe is None:
        hed_pipe = HedPipeline()
        hed_pipe.to("cpu")
    return hed_pipe(image)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli import hf_endpoint_url

# pretrained infos
pretrained_bloom_infos = {
    # bloom
    "bloom-560m": {
        "config": hf_endpoint_url("/bigscience/bloom-560m/resolve/main/config.json"),
        "tokenizer": hf_endpoint_url(
            "/bigscience/bloom-560m/resolve/main/tokenizer.json"
        ),
        "weight": hf_endpoint_url(
            "/bigscience/bloom-560m/resolve/main/pytorch_model.bin"
        ),
    },
    "bloom-3b": {
        "config": hf_endpoint_url("/bigscience/bloom-3b/resolve/main/config.json"),
        "tokenizer": hf_endpoint_url(
            "/bigscience/bloom-3b/resolve/main/tokenizer.json"
        ),
        "weight": hf_endpoint_url(
            "/bigscience/bloom-3b/resolve/main/pytorch_model.bin"
        ),
    },
    "bloom-7b1": {
        "config": hf_endpoint_url("/bigscience/bloom-7b1/resolve/main/config.json"),
        "tokenizer": hf_endpoint_url(
            "/bigscience/bloom-7b1/resolve/main/tokenizer.json"
        ),
        "weight": [
            hf_endpoint_url(
                "/bigscience/bloom-7b1/resolve/main/pytorch_model-{str(i).rjust(5, '0')}-of-00002.bin"
            )
            for i in range(1, 3)
        ],
    },
    "bloomz-560m": {
        "config": hf_endpoint_url("/bigscience/bloomz-560m/resolve/main/config.json"),
        "tokenizer": hf_endpoint_url(
            "/bigscience/bloomz-560m/resolve/main/tokenizer.json"
        ),
        "weight": hf_endpoint_url(
            "/bigscience/bloomz-560m/resolve/main/pytorch_model.bin"
        ),
    },
    "bloomz-3b": {
        "config": hf_endpoint_url("/bigscience/bloomz-3b/resolve/main/config.json"),
        "tokenizer": hf_endpoint_url(
            "/bigscience/bloomz-3b/resolve/main/tokenizer.json"
        ),
        "weight": hf_endpoint_url(
            "/bigscience/bloomz-3b/resolve/main/pytorch_model.bin"
        ),
    },
    "bloomz-7b1": {
        "config": hf_endpoint_url("/bigscience/bloomz-7b1/resolve/main/config.json"),
        "tokenizer": hf_endpoint_url(
            "/bigscience/bloomz-7b1/resolve/main/tokenizer.json"
        ),
        "weight": hf_endpoint_url(
            "/bigscience/bloomz-7b1/resolve/main/pytorch_model.bin"
        ),
    },
}

pretrained_bloom_extensions_infos = {}

import unitorch.cli.models.bloom.modeling
import unitorch.cli.models.bloom.processing
from unitorch.cli.models.bloom.modeling import (
    BloomForClassification,
    BloomForGeneration,
)
from unitorch.cli.models.bloom.processing import BloomProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

pretrained_peft_infos = {}

from unitorch.utils import is_diffusers_available

import unitorch.cli.models.peft.modeling_bloom
import unitorch.cli.models.peft.modeling_clip
import unitorch.cli.models.peft.modeling_llama
import unitorch.cli.models.peft.modeling_mistral

if is_diffusers_available():
    import unitorch.cli.models.peft.diffusers
from unitorch.cli.models.peft.modeling_bloom import (
    BloomLoraForClassification,
    BloomLoraForGeneration,
)
from unitorch.cli.models.peft.modeling_llama import (
    LlamaLoraForClassification,
    LlamaLoraForGeneration,
)
from unitorch.cli.models.peft.modeling_llava import (
    LlavaMistralClipLoraForClassification,
    LlavaMistralClipLoraForGeneration,
    LlavaLlamaSiglipLoraForGeneration,
)
from unitorch.cli.models.peft.modeling_mistral import (
    MistralLoraForClassification,
    MistralLoraForGeneration,
)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
import torch
import torchvision
import torch.nn as nn
import torch.nn.functional as F
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from diffusers import AutoencoderKL
from unitorch.models import (
    GenericModel,
    GenericOutputs,
    QuantizationConfig,
    QuantizationMixin,
)


class VAEForDiffusion(GenericModel):
    prefix_keys_in_state_dict = {
        "^encoder.*": "vae.",
        "^decoder.*": "vae.",
        "^post_quant_conv.*": "vae.",
        "^quant_conv.*": "vae.",
    }

    replace_keys_in_state_dict = {
        "\.query\.": ".to_q.",
        "\.key\.": ".to_k.",
        "\.value\.": ".to_v.",
        "\.proj_attn\.": ".to_out.0.",
    }

    def __init__(
        self,
        config_path: str,
        patch_size: Optional[int] = 32,
        stride: Optional[int] = 16,
    ):
        super().__init__()
        config_dict = json.load(open(config_path))
        self.vae = AutoencoderKL.from_config(config_dict)
        self.patch_size = patch_size
        self.stride = stride

    def patch_mse_loss(self, y_true, y_pred):
        y_true = y_true.unfold(2, self.patch_size, self.stride).unfold(
            3, self.patch_size, self.stride
        )
        y_pred = y_pred.unfold(2, self.patch_size, self.stride).unfold(
            3, self.patch_size, self.stride
        )
        y_true = y_true.contiguous().view(y_true.size(0), -1)
        y_pred = y_pred.contiguous().view(y_pred.size(0), -1)
        return F.mse_loss(y_pred, y_true, reduction="none").mean(1)

    def forward(self, pixel_values):
        latents = self.vae.encode(pixel_values).latent_dist.sample()
        outputs = self.vae.decode(latents).sample

        loss = self.patch_mse_loss(pixel_values, outputs).mean()
        return loss



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli import hf_endpoint_url

# pretrained infos
pretrained_mask2former_infos = {
    "mask2former-swin-tiny-ade-semantic": {
        "config": hf_endpoint_url(
            "/facebook/mask2former-swin-tiny-ade-semantic/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/facebook/mask2former-swin-tiny-ade-semantic/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/facebook/mask2former-swin-tiny-ade-semantic/resolve/main/pytorch_model.bin"
        ),
    },
    "mask2former-swin-base-ade-semantic": {
        "config": hf_endpoint_url(
            "/facebook/mask2former-swin-base-ade-semantic/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/facebook/mask2former-swin-base-ade-semantic/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/facebook/mask2former-swin-base-ade-semantic/resolve/main/pytorch_model.bin"
        ),
    },
    "mask2former-swin-large-ade-semantic": {
        "config": hf_endpoint_url(
            "/facebook/mask2former-swin-large-ade-semantic/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/facebook/mask2former-swin-large-ade-semantic/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/facebook/mask2former-swin-large-ade-semantic/resolve/main/pytorch_model.bin"
        ),
    },
}

import unitorch.cli.models.mask2former.modeling
import unitorch.cli.models.mask2former.processing
from unitorch.cli.models.mask2former.modeling import Mask2FormerForSegmentation
from unitorch.cli.models.mask2former.processing import Mask2FormerProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli import hf_endpoint_url

# pretrained infos
pretrained_chinese_clip_infos = {
    "chinese-clip-vit-base-patch16": {
        "config": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-base-patch16/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-base-patch16/resolve/main/preprocessor_config.json"
        ),
        "vocab": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-base-patch16/resolve/main/vocab.txt"
        ),
        "weight": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-base-patch16/resolve/main/pytorch_model.bin"
        ),
    },
    "chinese-clip-vit-large-patch14": {
        "config": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-large-patch14/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-large-patch14/resolve/main/preprocessor_config.json"
        ),
        "vocab": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-large-patch14/resolve/main/vocab.txt"
        ),
        "weight": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-large-patch14/resolve/main/pytorch_model.bin"
        ),
    },
    "chinese-clip-vit-huge-patch14": {
        "config": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-huge-patch14/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-huge-patch14/resolve/main/preprocessor_config.json"
        ),
        "vocab": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-huge-patch14/resolve/main/vocab.txt"
        ),
        "weight": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-huge-patch14/resolve/main/pytorch_model.bin"
        ),
    },
    "chinese-clip-vit-huge-patch14-336px": {
        "config": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-huge-patch14-336px/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-huge-patch14-336px/resolve/main/preprocessor_config.json"
        ),
        "vocab": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-huge-patch14-336px/resolve/main/vocab.txt"
        ),
        "weight": hf_endpoint_url(
            "/OFA-Sys/chinese-clip-vit-huge-patch14-336px/resolve/main/pytorch_model.bin"
        ),
    },
}

import unitorch.cli.models.chinese_clip.modeling
import unitorch.cli.models.chinese_clip.processing
from unitorch.cli.models.chinese_clip.modeling import (
    ChineseClipForPretrain,
    ChineseClipForClassification,
    ChineseClipForTextClassification,
    ChineseClipForImageClassification,
)
from unitorch.cli.models.chinese_clip.processing import ChineseClipProcessor
